"""
Примеры json-объектов для построения различных объектов.
"""
import os


_chinook_db_file = os.path.dirname(os.path.dirname(os.path.abspath(__file__))) + '/examples/chinook.db'


# пример описания фильтра для датафрейма в json
panda_filter = [
    {
        'col_name': 'city.city_id',
        'op': '>',
        'value': 12,
        'is_q': False,
    },
    {
        'is_q': True,
        'op': 'or',
        'filters': [
            {
                'col_name': 'address.city_id',
                'op': 'in',
                'value': [33, 35],
                'is_q': False,
            },
            {
                'col_name': 'city.city_id',
                'op': '==',
                'value': 22,
                'is_q': False,
            },
        ],
    },
]


test_conf = {
    #  словарь с конфигурациями источников
    'sources': {
        'sqlite': {
            'id': 'sqlite', # id источника

            # формат данных, по которому будет создаваться источник:
            # json/conn_str/expression - словарь, строка подключения либо выражение
            'from': 'json',
            'type': 'sqlite',
            'value': {
                'dbname': _chinook_db_file,
            },
        },
    },

    # словарь с конфигурациями запросов к БД
    'queries': {
        'sqlite': {
            'id': 'sqlite',  # id запроса
            'source': 'sqlite',  # id источника

            # формат данных, по которому будет создаваться запрос:
            # json/expression - словарь либо выражение
            'from': 'expression', # json/expression

            # этот запрос будет формироваться из выражения
            'value': 'sql_select("albums")',
        },
        'sqlite2': {
            'id': 'sqlite2',
            'source': 'sqlite',
            'from': 'expression',
            'value': 'sql_select("artists")',
        },
        'sqlite3': {
            'id': 'sqlite3',
            'source': 'sqlite',
            'from': 'json',
            'value': {
                'is_subquery': False,
                'base': 'albums',
                'tables': ['albums', 'artists', 'tracks'],
                'joins': [
                    {'l': 'albums', 'r': 'artists', 'j': 'inner', 'on': {'l': 'ArtistId', 'r': 'ArtistId'}},
                    {'l': 'albums', 'r': 'tracks', 'j': 'inner', 'on': {'l': 'AlbumId', 'r': 'AlbumId'}},
                ],
                'filters': [
                    {'is_complex': True, 'op': 'and', 'filters': [
                        {'is_complex': False, 'table': 'albums', 'field': 'AlbumId', 'op': '>', 'value': 100},
                        {'is_complex': False, 'table': 'albums', 'field': 'AlbumId', 'op': '<', 'value': 300},
                    ]},
                    #{'is_complex': False, 'table': 'albums', 'field': 'AlbumId', 'op': 'in', 'value': {
                    #    'is_subquery': True, 'base': 'albums', 'tables': ['albums'], 'values': [{'table': 'albums', 'field': 'AlbumId'}],
                    #}},
                    #{'is_complex': False, 'table': 'albums', 'field': 'ArtistId', 'op': 'in', 'value': [19, 21]},
                ],
                'values': [
                    #{'table': 'albums', 'field': 'Title'}, {'table': 'artists', 'field': 'Name'}
                ],
                'order_by': [
                    {'table': 'albums', 'fields': ['Title'], 'type': 'asc'},
                ],
                'limit': 100,
            },
        },
    },

    #  словарь с конфигурациями датафреймов
    'dataframes': {
        'sqlite': {
            'id': 'sqlite',  # id датафрейма

            # конфигурация основы для датафрейма
            'base': {
                #  тип основы - результат запроса к бд (query, указыается id запроса)
                #  либо выражение - expression, результат которого должен быть в свою очередь датафреймом

                # этот датафрейм будет строиться по результатам запроса
                'type': 'query',
                'source': 'sqlite',  # id источника - актуально при type=query
                'value': 'sqlite',  #  'значение' основания - id запроса либо строка выражения
            },
            # список соединений
            'joins': [
                {
                    'from': 'expression', # аналогично, формат/откуда берется датафрейм - другой датафрейм или выражение
                    'value': 'dataframe(sql_execute(create_connection("sqlite", "database=' + _chinook_db_file + '"), sql_select("artists")))',
                    'on': ['albums.ArtistId', 'artists.ArtistId', ],
                    'type': 'inner',
                },
            ],
            'unions': [],                #{
                #    'from': 'expression',  # значение приклеиваемого дф будет получено из выражения
                #    'value': 'dataframe(sql_execute(create_connection("sqlite", "database=' + _chinook_db_file + '"), sql_select("albums")))',
                    #'dataframe(test5["city.city_id"])',
                    #'from': 'dataframe',
                    #'value': 'sqlite2',
                #}ery',
                #'source': 'sqlite',
                #'value': 'sqlite2',
        },
        'sqlite3': {
            'id': 'sqlite3',
            'base': {
                'type': 'query',
                'source': 'sqlite',
                'value': 'sqlite3',
            },
            'filters': [
                #{'from': 'json', 'value': {'is_q': False, 'col_name': 'tracks.Milliseconds', 'op': '>', 'value': 200000}},
                {'from': 'expression', 'value': 'df_filter("tracks.Milliseconds", ">", 200000) & df_filter("tracks.Milliseconds", "<", 300000)'},
            ],
        },
    },
}

#  кофигурация компонентов отчёта
components_conf = {
    'table1': {
        'id': 'table1',
        #  тип компонента - таблица или график - table/figure
        'type': 'table',

        #  основа - получаемые данные
        'base': {
            #  откуда берутся данные - датафрейм/выражение
            'from': 'dataframe',  # dataframe/expression/json
            'value': 'sqlite3',
        },
        #  информация по компоновке
        'layout': {
            # Заголовок таблицы, если нужно
            'title_text': 'Таблица1',
            # список столбцов датафрейма, которые исключаются при выводе
            'to_exclude': [],

            # словарь для переименования столбцов;
            # формат {'название_существующего_столбца1': 'новое_название_столбца', ...}
            'aliases': {
                'artists.Name': 'Исполнитель',
                'albums.Title': 'Альбом',
            },
        },
    },
    'lines': {
        'id': 'lines',
        'type': 'figure',  # график
        'engine': 'plotly',  # "движок" - библиотека, которая будет исп. при построении графика
        'figure_type': 'line',  # тип графика
        'base': {
            'from': 'list',  # формат данных, в этом случае - набор отдельных линий
            'value': [
                {
                    'name': 'U2',
                    'x_from': 'expression',
                    'x': 'get_column(filter(sqlite3, df_filter("artists.Name", "==", "U2")), "tracks.TrackId")',
                    'y_from': 'column',
                    'y': {'df_from': 'expression', 'dataframe': 'filter(sqlite3, df_filter("artists.Name", "==", "U2"))', 'column': 'tracks.Bytes'},
                },
                {
                    'name': 'Soundgarden',
                    'x_from': 'column',
                    'x': {'df_from': 'expression', 'dataframe': 'filter(sqlite3, df_filter("artists.Name", "==", "Soundgarden"))', 'column': 'tracks.TrackId'},
                    'y_from': 'expression',
                    'y': 'get_column(filter(sqlite3, df_filter("artists.Name", "==", "Soundgarden")), "tracks.Bytes")',
                },
            ],
        },
        'layout': {
            'showlegend': True,
            'title_text': 'Линии',
            'title_font_size': 12,
        },
    },
    'lines2': {
        'id': 'lines2',
        'type': 'figure',
        'engine': 'plotly',
        'figure_type': 'line',
        'base': {
            'from': 'grouped',  # формат данных, в этом случае - набор отдельных линий
            'value': {'df_from': 'expression', 'dataframe': 'df_head(sqlite3, 50)'},
            'line_group': 'artists.Name',
            'x': 'tracks.TrackId',
            'y': 'tracks.Milliseconds',
        },
        'layout': {
            'showlegend': True,
            'title_text': 'Линии2',
            'title_font_size': 12,
        },
    },
    'pie': {
        'id': 'pie',
        'type': 'figure',
        'engine': 'plotly',
        'figure_type': 'pie',
        'base': {
            'from': 'list',
            'value': [
                {
                    'value': 'min(sqlite3["tracks.Milliseconds"]/1000)',
                    'name': 'Мин. длительность' 
                },
                {
                    'value': 'max(sqlite3["tracks.Milliseconds"]/1000)',
                    'name': 'Макс. длительность' 
                },
                {
                    'value': 'mean(sqlite3["tracks.Milliseconds"]/1000)',
                    'name': 'Средняя длительность' 
                },
            ],
        },
        'layout': {
            'showlegend': True,
            'title_text': 'Пирог1',
            'title_font_size': 30,
        },
    },
    'pie2': {
        'id': 'pie2',
        'type': 'figure',
        'engine': 'plotly',
        'figure_type': 'pie',
        'base': {
            'from': 'dataframe',
            'value': {'df_from': 'dataframe', 'dataframe': 'sqlite3'},
            'values': 'tracks.UnitPrice',
            'names': 'tracks.Name',
        },
        'layout': {
            'showlegend': True,
            'title_text': 'Пирог2',
            'title_font_size': 18,
        },
    },
    'bar': {
        'id': 'bar',
        'type': 'figure',
        'engine': 'plotly',
        'figure_type': 'bar',
        'base': {
            'from': 'list',
            'value': [
                {
                    'x_from': 'column',
                    'y_from': 'expression',
                    'x': {'df_from': 'expression', 'dataframe': 'sqlite3', 'column': "tracks.Name"},
                    'y': 'sqlite3["tracks.Milliseconds"]',
                    'name': 'Длительность',
                },
                {
                    'x_from': 'expression',
                    'y_from': 'expression',
                    'x': 'sqlite3["tracks.Name"]',
                    'y': 'sqlite3["tracks.Bytes"]',
                    'name': 'Bytes',
                },
            ],
            'barmode': 'relative',
        },
        'layout': {
            'showlegend': True,
            'title_text': 'Столбцы',
            'title_font_size': 18,
        },
    },
    'bar2': {
        'id': 'bar2',
        'type': 'figure',
        'engine': 'plotly',
        'figure_type': 'bar',
        'base': {
            'from': 'dataframe',
            'value': {'df_from': 'dataframe', 'dataframe': 'sqlite3'},
            'y': 'tracks.Name',
            'x': 'tracks.Milliseconds',
            'barmode': 'relative',
        },
        'layout': {
            'showlegend': True,
            'title_text': 'Столбцы',
            'title_font_size': 18,
        },
    },
}


document_conf = {
    #  список id компонентов, которые будут включены в pdf-документ
    'components': ['lines2', 'table1', 'pie', 'pie2', 'bar', 'bar2'],
    'orientation': 'landscape',
}


"""
Вычисляемое значение в контексте.
"""


class Variable:
    def __init__(self, name, expr, evaluator=None):
        self.name = name  # название
        self.expr = expr  # значение-выражение
        self.evaluated = False  # флаг вычисленного значения
        self.value = None  # вычисленное значение
        self._evaluator = evaluator  # объект-вычислитель (используется ReportBuilder)

    def evaluate(self):
        """
        Получение вычисленного значения.
        Возвращается значение, если оно уже было вычислено.
        В противном случае значение предварительно вычисляется.
        """
        if not self.evaluated:
            self.value = self._evaluator.evaluate(self.expr)
            self.evaluated = True
        return self.value

    def force_set(self, value):
        """
        Принудительное задание значения без вычисления.
        """
        self.value = value
        self.evaluated = True

    def __str__(self):
        return f'{self.name} = {self.value if self.evaluated else self.expr}'



"""
Функции работы со строками.
"""

from .base import StrFunc, MandatoryArg, NonMandatoryArg


def _endswith(string, sub):
    return string.endswith(sub)

def _startswith(string, sub):
    return string.startswith(sub)

def _lower(string):
    return string.lower()

def _upper(string):
    return string.upper()

def _concat(*strings):
    s = ''
    for i in strings:
        s += f'{i}'
    return s

def _join(string, substrings):
    return f'{string}'.join(substrings)


class Contains(StrFunc):
    """
    Содержит ли строка другую строку.
    """
    description = 'Проверка, содержит ли строка другую строку'
    args_description = [
        MandatoryArg('Строка1', 0, [str,]),
        MandatoryArg('Строка2', 1, [str,]),
    ]

    def _operation(self, *args):
        return args[1] in args[0]


class Endswith(StrFunc):
    """
    Заканчивается ли строка другой строкой.
    """
    description = 'Проверка, заканчивается ли строка другой строкой'
    args_description = [
        MandatoryArg('Строка1', 0, [str,]),
        MandatoryArg('Строка2', 1, [str,]),
    ]

    def _operation(self, *args):
        return args[0].endswith(args[1])


class Len(StrFunc):
    """
    Получение длины строки.
    """
    description = 'Получение длины строки'
    operation = len
    args_description = [
        MandatoryArg('Строка', 0),
    ]


class Lower(StrFunc):
    """
    Перевод строки в нижний регистр.
    """
    description = 'Перевод строки в нижний регистр'
    args_description = [
        MandatoryArg('Строка', 0, [str]),
    ]

    def _operation(self, *args):
        return args[0].lower()


class Upper(StrFunc):
    """
    Перевод строки в верхний регистр.
    """
    description = 'Перевод строки в верхний регистр'
    args_description = [
        MandatoryArg('Строка', 0, [str]),
    ]

    def _operation(self, *args):
        return args[0].upper()


class Concat(StrFunc):
    """
    Сложение строк.
    """
    description = 'Сложение строк'
    args_description = [
        MandatoryArg('Строка1', 0, [str]),
        MandatoryArg('Строка2', 1, [str]),
    ]

    def _operation(self, *args):
        result = ''
        for arg in args:
            result += arg
        return result


class Join(StrFunc):
    """
    Объединение строк по разделителю.
    """
    description = 'Объединение строк через разделитель'
    args_description = [
        MandatoryArg('Разделитель', 0, [str]),
        MandatoryArg('Строка1', 1, [str]),
        MandatoryArg('Строка2', 2, [str]),
    ]

    def _operation(self, *args):
        sep = args[0]
        subs = args[1:]
        return sep.join(subs)


class Startswith(StrFunc):
    """
    Начинается ли строка с другой строки.
    """
    description = 'Проверка, начинается ли строка другой строкой'
    args_description = [
        MandatoryArg('Строка1', 0, [str]),
        MandatoryArg('Строка2', 1, [str]),
    ]

    def _operation(self, *args):
        return args[0].startswith(args[1])



from datetime import datetime, date, time

from .base import DataframeFunc, MandatoryArg, NonMandatoryArg
from ..sources.conn import DataQueryResult
from ..dataframes import panda_builder


_DF = panda_builder.pd.DataFrame


class Join(DataframeFunc):
    """
    Функция соединения датафреймов.
    """
    description = 'Функция соединения датафреймов'
    args_description = [
        MandatoryArg('Датафрейм1', 0, [_DF]),
        MandatoryArg('Датафрейм2', 1, [_DF]),
        MandatoryArg('Поле первого датафрейма', 2, [str]),
        MandatoryArg('Поле второго датафрейма', 3, [str]),
        MandatoryArg('Тип объединения', 4, [str]),
    ]

    @classmethod
    def get_name(self):
        return 'df_join'

    def _operation(self, *args):
        df1, df2 = args[0], args[1]
        lo, ro = args[2], args[3]
        join_type = args[4]
        joined = panda_builder.join(df1, df2, [lo, ro], join_type)
        return joined


class Union(DataframeFunc):
    """
    Функция объединения датафреймов.
    """
    description = 'Функция объединения датафреймов'
    args_description = [
        MandatoryArg('Датафрейм1', 0, [_DF]),
        MandatoryArg('Датафрейм2', 1, [_DF]),
    ]

    @classmethod
    def get_name(self):
        return 'df_union'

    def _operation(self, *args):
        return panda_builder.union(list(args))


class Order(DataframeFunc):
    """
    Функция упорядочивания строк датафреймов.
    """
    description = 'Функция упорядочивания строк датафреймов'
    args_description = [
        MandatoryArg('Датафрейм1', 0, [_DF]),
        MandatoryArg('Поле', 1, [str]),
        NonMandatoryArg('Направление', 2, [str]),
    ]

    @classmethod
    def get_name(cls):
        return 'df_order'

    def _operation(self, *args):
        asc = True
        try:
            asc = args[2]
            asc = False if asc == 'desc' else True
        except IndexError:
            pass
        return panda_builder.order_by(args[0], args[1], asc=asc)


class DfFilterDict:
    """
    Обёртка для фильтров датафреймов в виде словарей.
    Используется для простого формирования и сложения фильтров в
    строках выражений.
    """
    def __init__(self, dff_dict):
        self._q = dff_dict

    @property
    def df_filter(self):
        return self._q

    def __and__(self, value):
        return DfFilterDict({
            'is_q': True,
            'op': 'and',
            'filters': [self.df_filter, value.df_filter],
        })

    def __or__(self, value):
        return DfFilterDict({
            'is_q': True,
            'op': 'or',
            'filters': [self.df_filter, value.df_filter],
        })

class DfFilter(DataframeFunc):
    """
    Функция создания фильтров для датафреймов.
    """
    description = 'Функция создания фильтров для датафреймов'
    args_description = [
        MandatoryArg('Название столбца', 1, [str]),
        MandatoryArg('Тип фильтра', 2, [str]),
        MandatoryArg('Значение', 3, [str, int, float, date, time, datetime, bool]),
    ]

    @classmethod
    def get_name(cls):
        return 'df_filter'

    def _operation(self, *args):
        return DfFilterDict({
            'is_q': False,
            'col_name': args[0],
            'op': args[1],
            'value': args[2],
        })


class Filter(DataframeFunc):
    """
    Функция применения фильтров датафреймов.
    """
    description = 'Функция применения фильтров датафреймов'
    args_description = [
        MandatoryArg('Датафрейм1', 0, [_DF]),
        MandatoryArg('Фильтр', 1, [DfFilterDict]),
    ]

    @classmethod
    def get_name(cls):
        return 'filter'

    def _operation(self, *args):
        filters = args[1]
        if not isinstance(filters, list):
            filters = [filters.df_filter]
        else:
            filters = [f.df_filter for f in filters]
        return panda_builder.filter(args[0], filters)


class DfQuery(DataframeFunc):
    """
    Функция выборки данных из датафрейма.
    """
    description = 'Функция выборки данных из датафрейма'
    args_description = [
        MandatoryArg('Датафрейм', 0, [_DF]),
        MandatoryArg('Строка запроса', 1, [str]),
    ]

    @classmethod
    def get_name(cls):
        return 'df_query'

    def _operation(self, *args):
        return args[0].query(args[1])


class AddColumn(DataframeFunc):
    """
    Функция добавления/изменения столбцов датафреймов.
    """
    description = 'Функция добавления/изменения столбцов датафреймов'
    args_description = [
        MandatoryArg('Датафрейм', 0, [_DF]),
        MandatoryArg('Название столбца', 1, [str]),
        MandatoryArg('Значения', 2),
    ]

    @classmethod
    def get_name(cls):
        return 'add_column'

    def _operation(self, *args):
        df, col_name, values = args[0], args[1], args[2]
        return panda_builder.add_column(df, col_name, values)


class GetColumn(DataframeFunc):
    """
    Функция получения столбца датафрейма.
    """
    description = 'Функция получения столбца датафрейма'
    args_description = [
        MandatoryArg('Датафрейм', 0, [_DF]),
        MandatoryArg('Название столбца', 1, [str]),
    ]

    @classmethod
    def get_name(cls):
        return 'get_column'

    def _operation(self, *args):
        return args[0][args[1]]


class Head(DataframeFunc):
    """
    Функция получения n первых строк датафрейма.
    """
    description = 'Функция получения n первых строк датафрейма'
    args_description = [
        MandatoryArg('Датафрейм', 0, [_DF]),
        MandatoryArg('Количество строк', 1, [int]),
    ]

    @classmethod
    def get_name(self):
        return 'df_head'

    def _operation(self, *args):
        return args[0].head(args[1])


class Tail(DataframeFunc):
    """
    Функция получения n последних строк датафрейма.
    """
    description = 'Функция получения n последних строк датафрейма'
    args_description = [
        MandatoryArg('Датафрейм', 0, [_DF]),
        MandatoryArg('Количество строк', 1, [int]),
    ]

    @classmethod
    def get_name(self):
        return 'df_tail'

    def _operation(self, *args):
        return args[0].tail(args[1])

class DfCount(DataframeFunc):
    """
    Функция получения количества строк датафрейма.
    """
    description = 'Функция получения количества строк датафрейма'
    args_description = [
        MandatoryArg('Датафрейм', 0, [_DF]),
    ]

    @classmethod
    def get_name(self):
        return 'df_count'

    def _operation(self, *args):
        df = args[0]
        return df.shape[0]


class CreateDataframe(DataframeFunc):
    """
    Функция создания датафрейма.
    """
    description = 'Функция создания датафреймов'
    args_description = [
        NonMandatoryArg('Данные', 0, [_DF, list, DataQueryResult]),
    ]

    @classmethod
    def get_name(cls):
        return 'dataframe'

    def _operation(self, *args):
        if not args:
            return panda_builder.new()
        data = args[0]
        if isinstance(data, DataQueryResult):
            return panda_builder.new(data.data, data.columns)
        if data:
            columns = data.pop(0)
            return panda_builder.new(data, columns)
        return panda_builder.new(data)



# словарь функций, где ключ - имя функции в нижнем регистре,
# значение - класс функции.
FUNCTIONS_DICT = {}

# аналогичный словарь операторов
OPERATORS_DICT = {}


def _collect_functions():
    """
    Сбор функций из подмодулей в словарь функций.
    """
    from .base import Func
    from . import (
        numbers,
        strings,
        date,
        logic,
        datastruct,
        typeconv,
        df,
        statistics,
        sql,
        conn,
    )

    fmodules = (
        numbers,
        strings,
        date,
        logic,
        datastruct,
        typeconv,
        df,
        statistics,
        sql,
        conn,
    )
    for fmod in fmodules:
        for f in dir(fmod):
            p = getattr(fmod, f)
            try:
                is_func = issubclass(p, Func)
            except TypeError:
                pass
            else:
                if is_func and p.get_description() != '__':
                    name = p.get_name()
                    FUNCTIONS_DICT[name] = p


def _collect_operators():
    """
    Сбор операторов в словарь операторов.
    """
    from .base import Func
    from . import operators
    for c in dir(operators):
        p = getattr(operators, c)
        try:
            is_func = issubclass(p, Func)
        except TypeError:
            pass
        else:
            if is_func and hasattr(p, 'sign'):
                OPERATORS_DICT[p.sign] = p


def collect_func_info():
    """
    Сбор информации по функциям.
    """
    res = {}
    for fname, f in FUNCTIONS_DICT.items():
        fgroup = f.get_group()
        if not fgroup in res:
            res[fgroup] = {}
        #res[fgroup][fname] = f.get_description()
        res[fgroup][fname] = f.get_func_info()
    return res


_collect_functions()
_collect_operators()



from .exceptions import NoSuchFieldError, TableHasNoFields


class Join:
    """
    Объединение таблиц.
    """

    join_type = 'INNER'

    def __init__(self, table1, table2, on):
        self._left = table1  # первая таблица
        self._right = table2  # вторая таблица

        """
        Список или кортеж, элементы которого - названия полей
        таблиц, по которым делается объединение;
        первый элемент - для левой таблицы, второй - для правой.
        """
        self._on = on

    def __str__(self):
        right_name, left_name = self._right._name, self._left._name
        left_on, right_on = self._on[0], self._on[1]
        if self._left.enquote or self._right.enquote:
            right_name, left_name = f'"{right_name}"', f'"{left_name}"'
            left_on, right_on = f'"{left_on}"', f'"{right_on}"'
        s = f'{self.__class__.join_type} JOIN {right_name} '
        s += f'ON {left_name}.{left_on} = {right_name}.{right_on}'
        return s


class LeftJoin(Join):
    join_type = 'LEFT OUTER'


class RightJoin(Join):
    join_type = 'RIGHT OUTER'


class SelectQuery:
    """
    Select-запрос.
    """

    def __init__(self, table, *fields):
        self._fields = []  # список возможных полей в запросе
        self._where = []  # список фильтров в запросе
        self._joins = []  # список объединений в запросе
        self._params = []  # список параметров для фильтров запроса
        self._table = table  # таблица
        self._values = []  # список запрашиваемых полей
        self._order_by = []
        self._limit = None
        self._is_mssql = False

        for f in table.get_verbose_names():
            self._fields.append(f)

        self.values(*fields)

    @property
    def requested(self):
        if self._values:
            values = [v[0] if v[1] is None else v[1] for v in self._values]
            return values
        return self._fields

    def join(self, *joins):
        """
        Добавление объединения/объединений в запрос.
        """
        for j in joins:
            right, left = j._right, j._left
            self._joins.append(j)

            if self._table == right:
                t = left
            else:
                t = right

            for f in t.get_verbose_names():
                if not f in self._fields:
                    self._fields.append(f)

        return self

    def where(self, *filters):
        """
        Добавление фильтров в запрос.
        """
        for f in filters:
            if type(f) is Q:
                if not f.is_empty:
                    self._where.append(f)
            else:
                self._where.append(Q(f))
            for p in f._params:
                self._params.append(p)

        return self

    @property
    def params(self):
        """
        Список параметров в фильтрах запроса.
        """
        return self._params

    def values(self, *values):
        for value in values:
            f = value[0]
            if f in self._fields:
                if f not in self._values:
                    self._values.append(value)
            else:
                raise NoSuchFieldError(f)
        return self

    def limit(self, n):
        try:
            n = int(n)
        except:
            pass
        else:
            self._limit = n
        return self

    def mssql_limit(self, n):
        self._is_mssql = True
        return self.limit(n)

    def order_by(self, *ords):
        for o in ords:
            fields, sorting = o[0], o[1].upper()
            for field in fields:
                if field in self._fields:
                    pass
                else:
                    raise NoSuchFieldError(field)
            assert sorting in ['ASC', 'DESC'], 'Значение должно быть равно ASC либо DESC'
            ordby = f'{", ".join(fields)} {sorting}'
            if ordby not in self._order_by:
                self._order_by.append(ordby)
        return self

    def _execute(self, cursor):
        q = str(self)
        params = self._params

        if params:
            cursor.execute(q, params)
        else:
            cursor.execute(q)

        return cursor.fetchall()

    def str_and_params(self):
        return (str(self), self._params)

    def __str__(self):
        if self._values:
            values = [v[0] for v in self._values]
            fields = f', '.join(values)
        else:
            fields = f', '.join(self._fields)

        has_limit = self._limit is not None

        q_beg = f'SELECT TOP {self._limit}' if has_limit and self._is_mssql else 'SELECT'
        q = f'{q_beg} {fields} FROM {self._table._name}'

        _joins = ''
        if self._joins:
            for j in self._joins:
                _joins += f'{str(j)} '

        if _joins != '':
            q += f' {_joins}'

        _where = ''
        if self._where:
            filters = [str(f) for f in self._where]
            _where = ' AND '.join(filters)

        if _where:
            q += f' WHERE {_where}'

        if self._order_by:
            ordby = f' ORDER BY {", ".join(self._order_by)}'
            q += ordby

        if has_limit and not self._is_mssql:
            q += f' LIMIT {self._limit}'

        return q.rstrip()


class Q:
    """
    Составной фильтр.
    """
    def __init__(self, *filters):
        self._filters = ['AND', ]
        self._params = []

        for f in filters:
            self._filters.append(f)
            for p in f._params:
                self._params.append(p)

    @property
    def is_empty(self):
        """
        Является фильтр пустым, т.е. не содержащим дургие фильтры.
        """
        return len(self._filters) < 2

    @property
    def params(self):
        """
        Список параметров в фильтрах запроса.
        """
        return self._params

    def _to_str(self, filters):
        op = f' {filters[0]} '
        res = []
        for i in filters[1:]:
            if type(i) is str:
                continue
            if type(i) is list:
                i = self._to_str(i)
                if i:
                    res.append(i)
            else:
                i = str(i)
                if i:
                    res.append(i)

        s = f'{op.join(res)}'
        if len(res) > 1:
            s = f'({s})'

        return s

    def __str__(self):
        return self._to_str(self._filters)

    def _add_q(self, value, disj=False):
        if type(value) is Q:
            if value.is_empty:
                return self
            if self.is_empty:
                return value

            q = Q()
            if disj:
                q._filters = ['OR', ]
            else:
                q._filters = ['AND', ]

            q._filters.append(self._filters)
            q._filters.append(value._filters)
            q._params += self._params
            q._params += value._params
            return q
        raise TypeError(f'Недопустимый операнд: {value}')

    def __and__(self, value):
        """
        Логическое умножение (AND) с другими составными фильтрами.
        """
        return self._add_q(value)

    def __or__(self, value):
        """
        Логическое сложение (OR) с другими составными фильтрами.
        """
        return self._add_q(value, disj=True)


class Filter:
    """
    Фильтр (WHERE в запросе).
    """

    def __init__(self, field_name, value, param_placeholder='%s'):
        self._field_name = field_name
        self._params = [value]
        self._param_placeholder = param_placeholder

    def __str__(self):
        return f"{self._field_name} {self.__class__.operator} {self._param_placeholder}"


class EqFilter(Filter):
    operator = '='


class NeqFilter(Filter):
    operator = '<>'


class GeFilter(Filter):
    operator = '>='


class GtFilter(Filter):
    operator = '>'


class LeFilter(Filter):
    operator = '<='


class LtFilter(Filter):
    operator = '<'


class ContainsFilter(Filter):
    operator = 'LIKE'

    def __init__(self, field_name, value, param_placeholder='%s'):
        self._field_name = field_name
        self._params = [f'%{value}%']
        self._param_placeholder = param_placeholder


class StartsWithFilter(Filter):
    operator = 'LIKE'

    def __init__(self, field_name, value, param_placeholder='%s'):
        self._field_name = field_name
        self._params = [f'{value}%']
        self._param_placeholder = param_placeholder


class EndsWithFilter(Filter):
    operator = 'LIKE'

    def __init__(self, field_name, value, param_placeholder='%s'):
        self._field_name = field_name
        self._params = [f'%{value}']
        self._param_placeholder = param_placeholder


class InFilter:
    """ IN """

    _not = False

    def __init__(self, field_name, values_range, param_placeholder='%s'):
        self._field_name = field_name
        self._not = False
        self._param_placeholder = param_placeholder

        if type(values_range) is SelectQuery:
            self._params = [p for p in values_range._params]
            self._subquery = values_range
        else:
            self._params = values_range
            self._subquery = None

    def __str__(self):
        operator = 'IN'
        if self.__class__._not == True:
            operator = 'NOT IN'

        if self._subquery is not None:
            return f'{self._field_name} {operator} ({str(self._subquery)})'
        else:
            placeholder = self._param_placeholder
            if self._params:
                if placeholder == '?':
                    params_list = ', '.join(placeholder*len(self._params))
                    return f'{self._field_name} {operator} ({params_list})'
                else:
                    params_list = ', '.join([placeholder]*len(self._params))
                return f'{self._field_name} {operator} ({placeholder})' % params_list
            else:
                return '1 = 1' if self._not else '1 = 2'

    def __neg__(self):
        self._not = not self._not
        return self


class NotInFilter(InFilter):
    _not = True


class Field:
    """
    Поле.
    Используется для фильтров в запросах.
    """

    def __init__(self, name, table):
        self._name = name
        self._table = table
        self._name_verbose = f'{table._name}.{name}'

    @property
    def table(self):
        return self._table

    def __eq__(self, value):
        return EqFilter(self._name_verbose, value)

    def __ge__(self, value):
        return GeFilter(self._name_verbose, value)

    def __gt__(self, value):
        return GtFilter(self._name_verbose, value)

    def __le__(self, value):
        return LeFilter(self._name_verbose, value)

    def __lt__(self, value):
        return LtFilter(self._name_verbose, value)

    def __ne__(self, value):
        return NeqFilter(self._name_verbose, value)

    def __mod__(self, value):
        return ContainsFilter(self._name_verbose, value)

    def __rshift__(self, value):
        return InFilter(self._name_verbose, value)


class Column:
    """
    Столбец.
    Используется для объединений.
    """

    def __init__(self, name, table):
        self._name = name
        self._table = table
        self._name_verbose = f'{table._name}.{name}'

    def _get_join(self, value, join_type):
        if type(value) is Column:
            field = value
            on = (self._name, field._name)

            if join_type == 'inner':
                j = Join
            elif join_type == 'left':
                j = LeftJoin
            elif join_type == 'right':
                j = RightJoin

            return j(self._table, field._table, on)
        else:
            raise TypeError(f'Недопустимый операнд: {value}')

    def __eq__(self, value):
        return self._get_join(value, 'inner')    

    def __lshift__(self, value):
        return self._get_join(value, 'left')

    def __rshift__(self, value):
        return self._get_join(value, 'right')


class ColumnSet:
    """
    Набор столбцов.
    """

    def __init__(self, columns):
        for col in columns:
            setattr(self, col._name, col)

    def __getitem__(self, key):
        return getattr(self, key)


class FieldSet:
    """
    Набор полей.
    """

    def __init__(self, fields):
        self._empty = len(fields) == 0
        for field in fields:
            setattr(self, field._name, field)

    def __getitem__(self, key):
        return getattr(self, key)

    @property
    def is_empty(self):
        return self._empty


class Table:
    """
    Таблица.
    """

    def __init__(self, name, fields, enquote=False):
        self._name = name
        self._fields_names = fields
        self.enquote = enquote  # брать ли в кавычки названия таблицы и полей

        _columns, _fields = [], []
        for field in fields:
            _fields.append(Field(field, self))
            _columns.append(Column(field, self))

        self.fields = FieldSet(_fields)
        self.columns = ColumnSet(_columns)
        if self.fields.is_empty:
            raise TableHasNoFields(self._name)

    @property
    def has_no_fields(self):
        return False

    def has_field(self, field):
        """
        Проверка, есть ли поле в таблице.
        """
        if field in self._fields_names:
            return True
        raise NoSuchFieldError(field)

    def get_fields_names(self):
        """
        Список полей таблицы.
        """
        return self._fields_names

    def get_verbose_names(self):
        """
        Список полных названий полей (с названием таблице в префиксе).
        """
        if self.enquote:
            verbose_names = [f'"{self._name}"."{field}"' for field in self._fields_names]
        else:    
            verbose_names = [f'{self._name}.{field}' for field in self._fields_names]
        return verbose_names

    def select(self, *fields):
        """
        Создание объекта простого select-запроса.
        """
        return SelectQuery(self, *fields)



class NoSuchFieldError(Exception):
    def __init__(self, field):
        self._field = field

    def __str__(self):
        return f'Поле {self._field} недоступно в запросе'


class TableHasNoFields(Exception):
    def __init__(self, table):
        self._table = table

    def __str__(self):
        return f'Таблица {self._table} не содержит полей либо не существует'



from .core.func import collect_func_info
from .core.dataframes import panda_builder
from .core.dataframes.utils import (
    get_dtypes,
    new_df,
    get_df_descriptor,
)
from .pdf import PdfReport
from .components.utils import apply_filters
from .components.pivot_chart import pivot_chart



import statistics

from .base import StatisticsFunc, MandatoryArg, NonMandatoryArg
from .. import pd


class Mean(StatisticsFunc):
    """
    Функция нахождения среднего значения.
    """
    description = 'Функция нахождения среднего значения'
    args_description = [
        MandatoryArg('Последовательность значений', 0, [list, pd.DataFrame]),
    ]

    def _operation(self, *args):
        seq = args[0]
        if isinstance(seq, list):
            mv = statistics.mean(seq)
        else:
            mv = seq.mean()
        return float(mv)


class Median(StatisticsFunc):
    """
    Функция нахождения медианы.
    """
    description = 'Функция нахождения медианы'
    args_description = [
        MandatoryArg('Последовательность значений', 0, [list, pd.DataFrame]),
    ]

    def _operation(self, *args):
        seq = args[0]
        if isinstance(seq, list):
            med = statistics.median(seq)
        else:
            med = seq.median()
        return float(med)


class Mode(StatisticsFunc):
    """
    Функция нахождения моды.
    """
    description = 'Функция нахождения моды'
    args_description = [
        MandatoryArg('Последовательность значений', 0, [list, pd.DataFrame]),
    ]

    def _operation(self, *args):
        seq = args[0]
        if isinstance(seq, list):
            m = statistics.mode(seq)
        else:
            m = seq.mode().loc(0)[0]
        return float(m)


class Min(StatisticsFunc):
    """
    Минимальное число в наборе.
    """
    description = 'Выбор наименьшего числа в наборе'
    args_description = [
        MandatoryArg('Набор чисел', 0, [list, pd.DataFrame]),
    ]

    def _operation(self, *args):
        seq = args[0]
        if isinstance(seq, list):
            return min(seq)
        return float(seq.min())


class Max(StatisticsFunc):
    """
    Максимальное число в наборе.
    """
    description = 'Выбор наибольшего числа в наборе'
    args_description = [
        MandatoryArg('Набор чисел', 0, [list, pd.DataFrame]),
    ]

    def _operation(self, *args):
        seq = args[0]
        if isinstance(seq, list):
            return max(seq)
        return float(seq.max())


class Stdev(StatisticsFunc):
    """
    Стандартное отклонение.
    """
    description = 'Функция вычисления стандартного отклонения'
    args_description = [
        MandatoryArg('Последовательность значений', 0, [list, pd.DataFrame]),
        NonMandatoryArg('Число степеней свободы', 1, [int]),
    ]

    def _operation(self, *args):
        seq = args[0]
        try:
            ddof = args[1]
        except IndexError:
            ddof = 1
        if isinstance(seq, list):
            seq = pd.Series(seq)
        return float(seq.std(ddof=ddof))


class Variance(StatisticsFunc):
    """
    Дисперсия.
    """
    description = 'Функция вычисления дисперсии'
    args_description = [
        MandatoryArg('Последовательность значений', 0, [list, pd.DataFrame]),
        NonMandatoryArg('Число степеней свободы', 1, [int]),
    ]

    def _operation(self, *args):
        seq = args[0]
        try:
            ddof = args[1]
        except IndexError:
            ddof = 1
        if isinstance(seq, list):
            seq = pd.Series(seq)
        return float(seq.var(ddof=ddof))



class ReportComponent:
    """
    Базовый класс для компонентов отчёта.
    """
    def __init__(self, id=None, build_from=None,
        base_value=None, title_text='', title_font_size=12,
        evaluator=None, **options):
        self.id = id
        self.type = None  # chart, table, pivot_table
        self._build_from = build_from
        self._base = base_value
        self.title_text = title_text
        self._title_font_size = title_font_size

        #  "вычислитель" - объект, который будет вычислять результаты
        #  в случае, когда base либо её части строятся из выражения ('expression');
        #  этим объектом на данный момент должен быть экз. ReportBuilder из noofa.builder.
        self._evaluator = evaluator

    @property
    def evaluator(self):
        return self._evaluator

    @evaluator.setter
    def evaluator(self, value):
        self._evaluator = value

    @property
    def build_from(self):
        """
        Формат, из которого строится основание для компонента.
        """
        return self._build_from

    @property
    def base(self):
        """
        Значение основания, из которого строится компонент.
        """
        return self._base



from pandas import DataFrame, Series
from datetime import datetime

from noofa.tests.base import NoofaFunctionTest
from noofa.core.func import logic as L


class TestLogicFunctions(NoofaFunctionTest):
    """
    Тестирование функций преобразования типов.
    """
    @classmethod
    def setUpClass(cls):
        cls.add_cases_from_list(_CASES)


_bad_args = (DataFrame(), Series((1,2)))

"""
Содержимое кортежей в списке _CASES:
(
    Функция,
    сабкейсы (кортеж из аргументов и ожидаемого результата),
    аргументы для проверки ошибки по типу аргумента,
    аргументы для проверки ошибки по кол-ву аргументов
)
"""
_CASES = [
    (L.BoolTrue, (((), bool, 'isinstance'), ((), True)), (), None),
    (L.BoolFalse, (((), bool, 'isinstance'), ((), False)), (), None),
    (
        L.Eq,
        (
            (('1', '1'), True), (('1', '2'), False),
            ((1, 1), True), ((1, 2), False),
            ((1.1, 1.1), True), ((1.2, 2.1), False),
            ((True, True), True), ((True, False), False),
            ((datetime(1970, 1, 1), datetime(1970, 1, 1)), True),
            ((datetime(1970, 1, 2), datetime(1970, 1, 1)), False),
            (([1, 1], [1, 1]), True), (([1, 2], [2, 1]), False),
        ),
        (_bad_args, ),
        (False, ),
    ),
    (
        L.Neq,
        (
            (('1', '1'), False), (('1', '2'), True),
            ((1, 1), False), ((1, 2), True),
            ((1.1, 1.1), False), ((1.2, 2.1), True),
            ((True, True), False), ((True, False), True),
            ((datetime(1970, 1, 1), datetime(1970, 1, 1)), False),
            ((datetime(1970, 1, 2), datetime(1970, 1, 1)), True),
            (([1, 1], [1, 1]), False), (([1, 2], [2, 1]), True),
        ),
        (_bad_args, ),
        (False, ),
    ),
    (
        L.Gt,
        (
            ((1, 1), False), ((2, 1), True),
            ((1.1, 1.1), False), ((2.2, 2.1), True),
            ((True, False), True), ((False, True), False),
            ((datetime(1970, 1, 2), datetime(1970, 1, 1)), True),
            ((datetime(1970, 1, 1), datetime(1970, 1, 2)), False),
        ),
        (_bad_args, ('1', '2'), ),
        (False, ),
    ),
    (
        L.Gte,
        (
            ((1, 1), True), ((2, 1), True), ((1, 2), False),
            ((1.1, 1.1), True), ((2.2, 2.1), True), ((2.2, 2.3), False),
            ((True, True), True), ((False, True), False), ((True, False), True),
            ((datetime(1970, 1, 2), datetime(1970, 1, 1)), True),
            ((datetime(1970, 1, 1), datetime(1970, 1, 1)), True),
            ((datetime(1970, 1, 1), datetime(1970, 1, 2)), False),
        ),
        (_bad_args, ('1', '2'), ),
        (False, ),
    ),
    (
        L.Lt,
        (
            ((1, 1), False), ((2, 3), True),
            ((1.1, 1.1), False), ((2.2, 2.3), True),
            ((True, False), False), ((False, True), True),
            ((datetime(1970, 1, 2), datetime(1970, 1, 1)), False),
            ((datetime(1970, 1, 1), datetime(1970, 1, 2)), True),
        ),
        (_bad_args, ('1', '2'), ),
        (False, ),
    ),
    (
        L.Lte,
        (
            ((1, 1), True), ((2, 1), False), ((1, 2), True),
            ((1.1, 1.1), True), ((2.2, 2.1), False), ((2.2, 2.3), True),
            ((True, True), True), ((False, True), True), ((True, False), False),
            ((datetime(1970, 1, 2), datetime(1970, 1, 1)), False),
            ((datetime(1970, 1, 1), datetime(1970, 1, 1)), True),
            ((datetime(1970, 1, 1), datetime(1970, 1, 2)), True),
        ),
        (_bad_args, ('1', '2'), ),
        (False, ),
    ),
    (L.If, (((True, 1, 2), 1), ((False, 1, 2), 2)), (), (False, )),
    (L.And, (((True, False), False), ((True, True), True)), (), (False, )),
    (L.Or, (((True, False), True), ((True, True), True), ((False, False), False)), (), (False, )),
    (L.In, (
        ((1, [1, 2]), True),
        (('1', '12'), True),
        ((10, [1, 2]), False),
        (('10', '12'), False),
        ), (), (False, )
    ),
    (L.Switch,
        (
            ((True, 0, (False, 2), (True, 1)), 1),
        ),
        (),
        (False, )
    )
]



"""
Функции преобразования типов данных.
"""
from .base import TypeconvFunc, MandatoryArg


class To_Str(TypeconvFunc):
    """
    Преобразование в строку.
    """
    description = 'Преобразование в строку'
    args_description = [
        MandatoryArg('Значение', 0, [int, float, str]),
    ]

    def _operation(self, *args):
        return str(args[0])

    @property
    def name(self):
        return 'to_str'


class To_Float(TypeconvFunc):
    """
    Преобразование в тип float.
    """
    description = 'Преобразование в тип float'
    args_description = [
        MandatoryArg('Значение', 0, [str, int, float]),
    ]

    def _operation(self, *args):
        return float(args[0])

    @property
    def name(self):
        return 'to_float'


class To_Int(TypeconvFunc):
    """
    Преобразование в тип int.
    """
    description = 'Преобразование в тип int'
    args_description = [
        MandatoryArg('Значение', 0, [str, int, float]),
    ]

    def _operation(self, *args):
        return int(args[0])

    @property
    def name(self):
        return 'to_int'



import plotly.express as px
import plotly.graph_objects as go
import numpy as np

from .base import ReportComponent
from .mixins import PlotlyMixin
from ..core.dataframes.panda_builder import pd


class ReportFigure(ReportComponent):
    """
    Базовый компонент-график. Не должен использоваться
    при создании компонентов. Компоненты создаются по
    экземплярам наследуемых классов, в которых должен быть реализован
    метод build - метод построения графика.
    """
    def __init__(self, x='', y='', showlegend=False, names='', values='',
        barmode='stack', labels='', line_group='', **options):
        super().__init__(**options)
        self.type = 'figure'
        self._showlegend = showlegend
        self._figure = None
        self._names = names
        self._values = values
        self._line_group = line_group
        self._x_col, self._y_col = x, y
        self._barmode = barmode
        self._labels = labels

    @property
    def figure(self):
        return self._figure

    def build(self):
        """
        Построение графика.
        """
        pass

    def to_bytes(self):
        """
        График в виде байт.
        График должен быть предварительно построен
        при помощи метода build.
        """
        return self.figure.to_image()

    def _update_layout(self):
        """
        Обновление компоновки/оформления графика.
        """
        self.figure.update_layout(
            showlegend=self._showlegend,
            title_text=self.title_text,
            title_font_size=self._title_font_size,
        )

    def _eval_xy(self, from_, value):
        """
        Получение значений для осей x либо y по значениям параметров конфиги.
        Используется при построении линейных графиков и столбчатых диаграмм.
        """
        evaluator = self.evaluator
        if from_ == 'expression':
            res = evaluator.evaluate(value)
        elif from_ == 'column':
            df_str, col_name, df_from = value['dataframe'], value['column'], value['df_from']
            if df_from == 'expression':
                _df = evaluator.evaluate(df_str)
            else:
                _df = evaluator.get_or_build_dataframe(df_str)
            res = _df[col_name]
        return res


class PlotlyLine(ReportFigure, PlotlyMixin):
    """
    График с линиями с использованием plotly.
    """
    def build(self):
        base, build_from = self.base, self.build_from
        data, fig = [], px.line()
        evaluator = self.evaluator
        if build_from == 'list':
            for i in base:
                x_from, y_from = i['x_from'], i['y_from']
                _x, _y = i['x'], i['y']
                x, y = self._eval_xy(x_from, _x), self._eval_xy(y_from, _y)
                data.append({
                    'x': x,
                    'y': y,
                    'name': i['name'],
                })
            fig = px.line()
            for d in data:
                x, y, name = d['x'], d['y'], d['name']
                x, y = _to_list(x), _to_list(y)
                fig.add_trace(go.Scatter(x=x, y=y, name=name))
        elif build_from == 'dataframe':
            df = evaluator.evaluate(base['dataframe'])
            fig = px.line(df, x=self._x_col, y=self._y_col)
        elif build_from == 'grouped':
            df_from, df_str = base['df_from'], base['dataframe']
            if df_from == 'expression':
                df = evaluator.evaluate(df_str)
            else:
                df = evaluator.get_or_build_dataframe(df_str)
            fig = px.line(
                df,
                x=self._x_col, y=self._y_col,
                color=self._line_group,
                line_group=self._line_group,
            )
        self._figure = fig
        self._update_layout()
        return self.figure


class PlotlyPie(ReportFigure, PlotlyMixin):
    """
    Круговая диаграмма с использованием plotly.
    """
    def using_agg(self, dataframe=None, groupby=[],
        on='', func='count', evaluator=None, **kwargs):
        df = evaluator.get_or_build_dataframe(dataframe)

        if groupby:
            agg_df = df.groupby(groupby).agg(func)
        else:
            agg_df = self._no_groupby(df, on, func)
            return px.pie(agg_df, names='index', values=func)

        names = list(agg_df.index)
        names_col = ', '.join(groupby)
        agg_df[names_col] = names
        agg_col_name = f'{func}({on})'
        agg_df[agg_col_name] = agg_df[on]

        agg_df = agg_df.replace([np.nan], [None])
        return px.pie(agg_df, names=names_col, values=agg_col_name)

    def build(self):
        base, build_from = self.base, self.build_from
        fig = px.pie()
        evaluator = self.evaluator
        if build_from == 'list':
            values, names = [], []
            for i in base:
                value = evaluator.evaluate(i['value'])
                values.append(value)
                names.append(i['name'])
            fig = px.pie(values=values, names=names)
        elif build_from == 'dataframe':
            df = base['dataframe']
            df = evaluator.evaluate(df)
            values, names = self._values, self._names
            fig = px.pie(df, values=values, names=names)
        elif build_from == 'agg':
            fig = self.using_agg(evaluator=self.evaluator, **base)

        self._figure = fig
        self._update_layout()
        return self.figure


class PlotlyBar(ReportFigure, PlotlyMixin):
    """
    Столбчатая диаграмма с использованием plotly.
    """
    _orientation = 'v'

    @property
    def orientation(self):
        return self.__class__._orientation

    @property
    def is_vertical(self):
        return self.orientation == 'v'

    def using_agg(self, dataframe=None, groupby=[],
        on='', func='count', legend=[], evaluator=None, **kwargs):
        df = evaluator.get_or_build_dataframe(dataframe)
        kw = {'orientation': self.orientation}

        if groupby:
            if legend:
                return self._using_legend(df, legend, groupby, on, func)

            df = df.groupby(groupby).agg({on: func})
        else:
            df = self._no_groupby(df, on, func)
            x, y = ('index', func) if self.is_vertical else (func, 'index')
            return px.bar(df, x=x, y=y, **kw)

        df, indices_col = self._append_indices(df, groupby)
        agg_col_name = f'{func}({on})'
        df[agg_col_name] = df[on]

        x, y = indices_col, agg_col_name
        if self.orientation == 'h':
             x, y = y, x
        df = df.replace([np.nan], [None])
        return px.bar(df, x=x, y=y, **kw)

    def build(self):
        base, build_from = self.base, self.build_from
        orientation = self.orientation
        if build_from == 'list':
            data, fig = [], px.bar(orientation=orientation)
            for i in base:
                x_from, y_from = i['x_from'], i['y_from']
                _x, _y = i['x'], i['y']
                x, y = self._eval_xy(x_from, _x), self._eval_xy(y_from, _y)
                data.append({
                    'x': x,
                    'y': y,
                    'name': i['name'],
                })
            for d in data:
                x, y = _to_list(d['x']), _to_list(d['y'])
                fig.add_trace(go.Bar(x=x, y=y, name=d['name']))
        elif build_from == 'dataframe':
            df_str, x_col, y_col = base['dataframe'], self._x_col, self._y_col
            df = self.evaluator.evaluate(df_str)
            fig = px.bar(df, x=x_col, y=y_col, orientation=orientation)
        elif build_from == 'agg':
            fig = self.using_agg(evaluator=self.evaluator, **base)

        self._figure = fig
        fig.update_layout(barmode=self._barmode)
        self._update_layout()
        return self.figure

    def _append_indices(self, df, groupby):
        indices = list(df.index)
        if len(groupby) > 1:
            try:
                indices = [', '.join(i) for i in indices]
            except TypeError:
                _ = []
                for ind in indices:
                    _.append(', '.join([str(i) for i in ind]))
                indices = _
        df = df.reset_index()
        indices_col = ', '.join(groupby)
        df[indices_col] = indices
        return df, indices_col

    def _using_legend(self, df, legend, groupby, on, func):
        groupby_col = ', '.join(groupby)
        for n, l in enumerate(groupby):
            if n == 0:
                df[groupby_col] = df[l].astype('str')
                continue
            df[groupby_col] += ', ' + df[l].astype('str')

        legend_col_name = ', '.join(legend)
        for n, l in enumerate(legend):
            if n == 0:
                df[legend_col_name] = df[l].astype('str')
                continue
            df[legend_col_name] += ', ' + df[l].astype('str')

        distinct_groups = []
        _ = df.drop_duplicates(subset=groupby)[groupby]
        for r in _.to_dict(orient='records'):
            dg = ''
            for n, g in enumerate(groupby):
                if n == 0:
                    dg = r[g]
                    continue
                dg += f', {r[g]}'
            distinct_groups.append(dg)

        distinct_legends = []
        _ = df.drop_duplicates(subset=legend)[legend]
        for r in _.to_dict(orient='records'):
            dl = ''
            for n, l in enumerate(legend):
                if n == 0:
                    dl = r[l]
                    continue
                dl += f', {r[l]}'
            distinct_legends.append(dl)

        values = []
        legend_labels = []
        groups_labels = []
        for dg in distinct_groups:
            for dl in distinct_legends:
                q = f"`{groupby_col}` == '{dg}' & `{legend_col_name}` == '{dl}'"
                _ = df.query(q)
                groups_labels.append(dg)
                values.append(_.agg({on: func})[0])
                legend_labels.append(dl)

        func_col = f'{func}({on})'
        df_dict = {
            groupby_col: groups_labels,
            func_col: values,
            legend_col_name: legend_labels,
        }
        df = pd.DataFrame(df_dict)
        x, y = (groupby_col, func_col) if self.is_vertical else (func_col, groupby_col)
        df = df.replace([np.nan], [None])
        return px.bar(df, x=x, y=y, color=legend_col_name, orientation=self.orientation)


class PlotlyHbar(PlotlyBar, PlotlyMixin):
    _orientation = 'h'


FIGURES = {
    'plotly': {
        'line': PlotlyLine,
        'pie': PlotlyPie,
        'bar': PlotlyBar,
        'hbar': PlotlyHbar,
    },
}


def _to_list(data):
    """
    Преобразование pandas.Series и датафреймов pandas,
    содержащих один столбец, в list.
    """
    if isinstance(data, list):
        return data
    elif isinstance(data, pd.Series):
        return data.to_list()
    elif isinstance(data, pd.DataFrame):
        cols = data.columns
        if len(cols) == 1:
            return data[cols[0]].to_list()
    return []



from pandas import Series

from .values import Variable
from .base import Func, MandatoryArg, NonMandatoryArg
from .errors import (
    InterpreterContextError, 
    ValueNotInContextError,
    RecursiveEvaluationError,
)


class Context:
    """
    Контекст интерпретатора.
    """
    def __init__(self, **kwargs):
        self.global_context = {**kwargs}  # глобальный контекст (весь отчёт)
        self.local_context = {}  #  локальный контекст (в случае выч. значения в строке или столбце)
        self._current_context = self.global_context
        self._using_local = False
        self._values = {}  # вычисляемые значения

        #  используем "стэк" для контроля рекурсивных вычислений
        self._values_stack = []

    def add_values(self, values_list, evaluator):
        """
        Добавление значений в контекст.
        values_list - список словарей с ключами name и value,
        evaluator - объект-вычислитель.
        """
        for v in values_list:
            name, value = v['name'], v['value']
            self._values[name] = Variable(name, value, evaluator)

    def get_value(self, name):
        """
        Получение значения.
        """
        #  если значение уже вычисляется и какое-либо из значений, от которых
        #  оно зависит, требуют вычисления этого же значения, то выбрасываем ошибку по рекурсии
        if name in self._values_stack:
            self._values_stack.clear()  # очищаем "стэк" в случае ошибки по рекурсии
            raise RecursiveEvaluationError(name)
        try:
            value_obj = self._values[name]
        except KeyError:
            raise ValueNotInContextError(name)

        #  добавляем название значения в "стэк"
        self._values_stack.append(name)

        value = value_obj.evaluate()

        #  при успешном получении значения убираем название из "стэка"
        self._values_stack.pop()

        return value

    def update_value(self, name, value):
        """
        Обновление значения.
        """
        if name in self._values:
            self._values[name].force_set(value)

    def clear_local(self):
        self.local_context = {}

    def switch_to_global(self):
        """
        Переключение контекста на глобальный.
        """
        self._using_local = False
        self._current_context = self.global_context

    def switch_to_local(self):
        """
        Переключение контекста на локальный.
        """
        self._using_local = True
        self._current_context = self.local_context

    def add_to_global(self, key, value):
        """
        Добавление в глобальный контекст.
        """
        self.global_context[key] = value

    def add_to_local(self, key, value):
        """
        Добавление в локальный контекст.
        """
        self.local_context[key] = value

    def add(self, key, value):
        """
        Добавление в текущий контекст.
        """
        self._current_context[key] = value

    def get(self, key):
        """
        Получение значения из текущ. либо глобального контекста.
        """
        try:
            return self._current_context[key]
        except KeyError:
            try:
                return self.global_context[key]
            except KeyError:
                raise InterpreterContextError(key)

    def remove(self, key):
        """
        Удаление из текущего контекста.
        """
        self._current_context.pop(key)


class GetValue(Func):
    """
    Функция получения значения из контекста интерпретатора.
    Используется на уровне интерпретатора для получения вычисляемых значений.
    """
    group = 'context'
    description = 'Функция контекста'
    args_description = [
        MandatoryArg('context', 0),
        MandatoryArg('var', 1),
    ]

    @classmethod
    def get_name(cls):
        return 'get_value'

    def _operation(self, *args):
        ctx, name = args[0], args[1]
        return ctx.get_value(name)


class GetFromContext(Func):
    """
    Функция получения значения из контекста интерпретатора.
    Используется на уровне интерпретатора для получения датафреймов.
    """
    group = 'context'
    description = 'Функция контекста'
    args_description = [
        MandatoryArg('context', 0),
        MandatoryArg('var', 1),
    ]

    @classmethod
    def get_name(cls):
        return '_getfromcontext'

    def _operation(self, *args):
        return args[0].get(args[1])


class GetSlice(Func):
    """
    Функция получения столбца датафрейма.
    Используется на уровне интерпретатора.
    """
    group = 'context'
    description = 'Функция контекста'
    args_description = [
        MandatoryArg('obj', 0),
        MandatoryArg('col', 1),
    ]

    @classmethod
    def get_name(cls):
        return '_getslice'

    def _operation(self, *args):
        args = list(args)
        obj, key = args.pop(0), args[0]

        # в случае применения функции apply интерпретатора объектом может быть словарь
        if isinstance(obj, dict):
            result = obj[key[0]]
        else:
            result = obj[key]
        return result


class GetConnection(Func):
    """
    Функция получения соединения с источником.
    Используется на уровне интерпретатора.
    """
    group = 'context'
    description = 'Функция контекста'
    args_description = [
        MandatoryArg('context', 0),
        MandatoryArg('source_id', 1),
    ]

    @classmethod
    def get_name(cls):
        return 'connection'

    def _operation(self, *args):
        return args[0][args[1]]


_context_funcs = {
    GetFromContext.get_name(): GetFromContext,
    GetSlice.get_name(): GetSlice,
    GetConnection.get_name(): GetConnection,
    GetValue.get_name(): GetValue,
}



from noofa.tests.base import NoofaSuite, NoofaRunner
from noofa.tests.components.dataschema import TestDataschema
from noofa.tests.components.components import TestComponentsSchema


def components_suite():
    suite = NoofaSuite()
    suite.add(TestDataschema)
    suite.add(TestComponentsSchema)
    return suite


def run_tests():
    suite = components_suite()
    runner = NoofaRunner()
    runner.run(suite)



def apply_filters(df, filters):
    complex_filter = None
    for col, filters_conf in filters.items():
        type_, values = filters_conf['type'], filters_conf['values']
        dtype = str(df[col].dtype)
        is_num = dtype.startswith('int') or dtype.startswith('float')
        if is_num or dtype == 'bool':
            values = _conv(values, dtype, is_num)

        f = df[col].isin(values)
        if type_ == 'exclude':
            f = ~f
        if complex_filter is None:
            complex_filter = f
        else:
            complex_filter &= f

    if complex_filter is not None:
        return df[complex_filter]

    return df


def _conv(seq, dtype, is_num=False):
    def _bool(value):
        value = value.lower()
        if value == 'true':
            return True
        elif value == 'false':
            return False

    def _num(value):
        try:
            return float(value)
        except:
            return None

    if dtype == 'bool':
        f = _bool
    if is_num:
        f = _num

    converted_values = [i for i in map(f, seq) if i is not None]
    return converted_values



from pandas import DataFrame, Series
from datetime import datetime, time, date

from noofa.tests.base import NoofaFunctionTest
from noofa.core.func import statistics as S


class TestStatisticsFunctions(NoofaFunctionTest):
    """
    Тестирование статистических функций.
    """
    @classmethod
    def setUpClass(cls):
        cls.add_cases_from_list(_CASES)


_dt = datetime.now()
_bad_args = (_dt, _dt.date(), _dt.time(), 1, 3.14, '', False)


"""
Содержимое кортежей в списке _CASES:
(
    Функция,
    сабкейсы (кортеж из аргументов и ожидаемого результата),
    аргументы для проверки ошибки по типу аргумента,
    аргументы для проверки ошибки по кол-ву аргументов
)
"""
_CASES = [
    (S.Mean, (), _bad_args, ()),
    (S.Median, (), _bad_args, ()),
    (S.Mode, (), _bad_args, ()),
    (S.Min, (), _bad_args, ()),
    (S.Max, (), _bad_args, ()),
    (S.Stdev, (), _bad_args, ()),
    (S.Variance, (), _bad_args, ()),
]



"""
Структуры данных.
"""

from .base import DatastructFunc, NonMandatoryArg, MandatoryArg
from ..dataframes.panda_builder import pd


class List(DatastructFunc):
    """
    Функция создания списка элементов.
    """
    description = 'Функция создания списка элементов'
    args_description = [
        NonMandatoryArg('Элемент1', 0),
        NonMandatoryArg('Элемент2', 1),
        NonMandatoryArg('Элемент n', 2),
    ]

    def _operation(self, *args):
        return [arg for arg in args]


class ToList(DatastructFunc):
    """
    Функция преобразования в список элементов.
    """
    description = 'Функция преобразования в список элементов'
    args_description = [
        MandatoryArg('Элемент', 0, [pd.DataFrame, pd.Series]),
    ]

    @classmethod
    def get_name(cls):
        return 'to_list'

    def _operation(self, *args):
        arg = args[0]
        if isinstance(arg, pd.DataFrame):
            cols = arg.columns
            if len(cols) == 1:
                return arg[cols[0]].to_list()
        elif isinstance(arg, pd.Series):
            return arg.to_list()
        return []



from pandas import DataFrame

from noofa.tests.base import NoofaTest, NoofaInterpreterTest
from noofa.core.func import operators as O
from noofa.core.func.errors import (
    NotEnoughArguments,
    ArgumentTypeError,
    ExpressionParsingError,
    ExpressionSyntaxError,
    InterpreterContextError,
)


class TestInterpreter(NoofaInterpreterTest):
    """
    Тестирование интерпретатора выражений.
    """
    @classmethod
    def setUpClass(cls):
        cls.set_up_interpreter()

    def test_int(self):
        self.assertEqual(self.evaluate('5'), 5)

    def test_float(self):
        self.assertEqual(self.evaluate('5.6'), 5.6)
        self.assertEqual(self.evaluate('.6'), 0.6)

    def test_negative_number(self):
        self.assertEqual(self.evaluate('-5'), -5)
        self.assertEqual(self.evaluate('(-5)'), -5)

    def test_str(self):
        self.assertEqual(self.evaluate('"hello"'), 'hello')
        self.assertEqual(self.evaluate("'hello'"), 'hello')

    def test_and_or_priority(self):
        self.assertFalse(self.evaluate('false() | true() & false()'))

    def test_eq_neq_and_priority(self):
        self.assertTrue(self.evaluate('1 == 1 & 1 != 2'))

    def test_gt_gte_and_priority(self):
        self.assertFalse(self.evaluate('1 >= 2 & 2 > 1'))

    def test_lt_lte_and_priority(self):
        self.assertFalse(self.evaluate('1 <= 2 & 2 < 1'))

    def test_gt_gte_lt_lte_priority(self):
        self.assertTrue(self.evaluate('1 <= 2 & 2 > 1'))
        self.assertTrue(self.evaluate('2 >= 1 & 1 < 2'))

    def test_arithm_comparison_priority(self):
        self.assertTrue(self.evaluate('2 * 2 >= 3 + 1'))
        self.assertTrue(self.evaluate('6 - 3 <= 4 - 1'))

    def test_arithm_priority(self):
        self.assertEqual(self.evaluate('1 + 1 - 4'), -2)
        self.assertEqual(self.evaluate('1 - 4 + 1'), -2)
        self.assertEqual(self.evaluate('2 + 3 * 5'), 17)
        self.assertEqual(self.evaluate('(2 + 3) * 5'), 25)
        self.assertEqual(self.evaluate('2 + 3 / 5'), 2.6)
        self.assertEqual(self.evaluate('(2 + 3) / 5'), 1)
        self.assertEqual(self.evaluate('2 * 5 + 3'), 13)
        self.assertEqual(self.evaluate('6 * 2 / 4'), 3)
        self.assertEqual(self.evaluate('6 / 2 * 4'), 12)

    def test_value_not_found(self):
        self.assertRaises(InterpreterContextError, self.evaluate, 'non_existent_value')

    def test_function_not_found(self):
        self.assertRaises(Exception, self.evaluate, 'non_existent_function()')

    def test_syntax(self):
        cases = [
            '1..2',
            '==1',
            '!=1',
            '>=1',
            '>1',
            '<1',
            '<=1',
            '*1',
            '/1',
            '**',
            '//',
            '()',
        ]
        for case in cases:
            with self.subTest(i=case):
                self.assertRaises(
                    ExpressionSyntaxError,
                    self.evaluate,
                    case
                )

    def test_parsing_errors(self):
        cases = [
            ',',
            '!',
            '?',
            '"',
            '1 = 1',
        ]
        for case in cases:
            with self.subTest(i=case):
                self.assertRaises(
                    ExpressionParsingError,
                    self.evaluate,
                    case
                )

    def test_parethesis(self):
        self.assertEqual(self.evaluate('(1 + 2) * 3'), 9)

    def test_no_opening_parenthesis(self):
        self.assertRaises(ExpressionParsingError, self.evaluate, '1 + 2)')

    def test_no_closing_parenthesis(self):
        self.assertRaises(ExpressionParsingError, self.evaluate, '(1 + 2')


class TestOperators(NoofaInterpreterTest):
    """
    Тестирование операторов интерпретатора выражений.
    """
    @classmethod
    def setUpClass(cls):
        cls.set_up_interpreter()

    def test_add(self):
        self.assertEqual(self.evaluate('1 + 1'), 2)

    def test_add_df(self):
        self.assertIsInstance(self.evaluate('dataframe() + dataframe()'), DataFrame)

    def test_subtract(self):
        self.assertEqual(self.evaluate('1 - 1'), 0)

    def test_negative_number(self):
        self.assertEqual(self.evaluate('-1'), -1)
        self.assertEqual(self.evaluate('-1 - 1'), -2)

    def test_multiply(self):
        self.assertEqual(self.evaluate('2 * 3'), 6)

    def test_divide(self):
        self.assertEqual(self.evaluate('6 / 3'), 2)

    def test_zero_division(self):
        self.assertRaises(ZeroDivisionError, self.evaluate, '6 / 0')

    def test_gt(self):
        self.assertTrue(self.evaluate('2 > 1'))
        self.assertFalse(self.evaluate('1 > 2'))

    def test_gte(self):
        self.assertTrue(self.evaluate('2 >= 1'))
        self.assertTrue(self.evaluate('2 >= 2'))
        self.assertFalse(self.evaluate('1 >= 2'))

    def test_lt(self):
        self.assertTrue(self.evaluate('1 < 2'))
        self.assertFalse(self.evaluate('2 < 1'))

    def test_lte(self):
        self.assertTrue(self.evaluate('1 <= 2'))
        self.assertTrue(self.evaluate('1 <= 1'))
        self.assertFalse(self.evaluate('2 <= 1'))

    def test_eq(self):
        self.assertTrue(self.evaluate('1 == 1'))
        self.assertFalse(self.evaluate('1 == 2'))

    def test_neq(self):
        self.assertFalse(self.evaluate('1 != 1'))
        self.assertTrue(self.evaluate('1 != 2'))

    def test_and(self):
        self.assertTrue(self.evaluate('true() & true()'))
        self.assertFalse(self.evaluate('true() & false()'))
        self.assertFalse(self.evaluate('false() & false()'))

    def test_or(self):
        self.assertTrue(self.evaluate('true() | true()'))
        self.assertTrue(self.evaluate('true() | false()'))
        self.assertFalse(self.evaluate('false() | false()'))


class TestOperatorsClasses(NoofaTest):
    """
    Тестирование классов операторов.
    """
    def test_add(self):
        op = O.Add(1, 2)
        self.assertEqual(op(), 3)

    def test_add_raises_nea(self):
        op = O.Add(2)
        self.assertRaises(NotEnoughArguments, op)

    def test_add_raises_ate(self):
        op = O.Add(2, None)
        self.assertRaises(ArgumentTypeError, op)

    def test_add_df(self):
        op = O.Add(DataFrame(), DataFrame())
        self.assertIsInstance(op(), DataFrame)

    def test_subtract(self):
        op = O.Subtract(3, 1)
        self.assertEqual(op(), 2)

    def test_subtract_raises_nea(self):
        op = O.Subtract(2)
        self.assertRaises(NotEnoughArguments, op)

    def test_subtract_raises_ate(self):
        op = O.Subtract(2, None)
        self.assertRaises(ArgumentTypeError, op)

    def test_subtract_returns_negative(self):
        op = O.Subtract(None, 2)
        self.assertEqual(op(), -2)

    def test_mult(self):
        op = O.Multiply(3, 2)
        self.assertEqual(op(), 6)

    def test_mult_raises_nea(self):
        op = O.Multiply()
        self.assertRaises(NotEnoughArguments, op)

    def test_mult_raises_ate(self):
        op = O.Multiply(2, DataFrame())
        op2 = O.Multiply(2, '2')
        self.assertRaises(ArgumentTypeError, op)
        self.assertRaises(ArgumentTypeError, op2)

    def test_divide(self):
        op = O.Divide(6, 2)
        self.assertEqual(op(), 3)

    def test_divide_raises_nea(self):
        op = O.Divide(6)
        self.assertRaises(NotEnoughArguments, op)

    def test_divide_raises_ate(self):
        op = O.Divide(2, DataFrame())
        op2 = O.Divide(2, '2')
        self.assertRaises(ArgumentTypeError, op)
        self.assertRaises(ArgumentTypeError, op2)

    def test_gt(self):
        op1 = O.IsGt(2, 1)
        op2 = O.IsGt(1, 2)
        self.assertTrue(op1())
        self.assertFalse(op2())

    def test_gt_raises_nea(self):
        op = O.IsGt(6)
        self.assertRaises(NotEnoughArguments, op)

    def test_gt_raises_ate(self):
        op1 = O.IsGt(DataFrame(), 2)
        op2 = O.IsGt(2, '2')
        self.assertRaises(ArgumentTypeError, op1)
        self.assertRaises(ArgumentTypeError, op2)

    def test_gte(self):
        op1 = O.IsGte(2, 1)
        op2 = O.IsGte(1, 2)
        op3 = O.IsGte(1, 1)
        self.assertTrue(op1())
        self.assertFalse(op2())
        self.assertTrue(op3())

    def test_gte_raises_nea(self):
        op = O.IsGte(6)
        self.assertRaises(NotEnoughArguments, op)

    def test_gte_raises_ate(self):
        op1 = O.IsGte(DataFrame(), 2)
        op2 = O.IsGte(2, '2')
        self.assertRaises(ArgumentTypeError, op1)
        self.assertRaises(ArgumentTypeError, op2)

    def test_lt(self):
        op1 = O.IsLt(1, 2)
        op2 = O.IsLt(2, 1)
        self.assertTrue(op1())
        self.assertFalse(op2())

    def test_lt_raises_nea(self):
        op = O.IsLt(6)
        self.assertRaises(NotEnoughArguments, op)

    def test_lt_raises_ate(self):
        op1 = O.IsLt(DataFrame(), 2)
        op2 = O.IsLt(2, '2')
        self.assertRaises(ArgumentTypeError, op1)
        self.assertRaises(ArgumentTypeError, op2)

    def test_lte(self):
        op1 = O.IsLte(1, 2)
        op2 = O.IsLte(2, 1)
        op3 = O.IsLte(1, 1)
        self.assertTrue(op1())
        self.assertFalse(op2())
        self.assertTrue(op3())

    def test_lte_raises_nea(self):
        op = O.IsLte(6)
        self.assertRaises(NotEnoughArguments, op)

    def test_lte_raises_ate(self):
        op1 = O.IsLte(DataFrame(), 2)
        op2 = O.IsLte(2, '2')
        self.assertRaises(ArgumentTypeError, op1)
        self.assertRaises(ArgumentTypeError, op2)

    def test_eq(self):
        op1 = O.IsEq(1, 2)
        op2 = O.IsEq(1, 1)
        self.assertFalse(op1())
        self.assertTrue(op2())

    def test_eq_raises_nea(self):
        op = O.IsEq(1)
        self.assertRaises(NotEnoughArguments, op)

    def test_eq_raises_ate(self):
        op = O.IsEq(DataFrame(), 2)
        self.assertRaises(ArgumentTypeError, op)

    def test_neq(self):
        op1 = O.IsNeq(1, 2)
        op2 = O.IsNeq(1, 1)
        self.assertTrue(op1())
        self.assertFalse(op2())

    def test_neq_raises_nea(self):
        op = O.IsNeq(1)
        self.assertRaises(NotEnoughArguments, op)

    def test_neq_raises_ate(self):
        op = O.IsNeq(DataFrame(), 2)
        self.assertRaises(ArgumentTypeError, op)

    def test_and(self):
        op1 = O.And(True, False)
        op2 = O.And(True, True)
        op3 = O.And(False, False)
        self.assertFalse(op1())
        self.assertTrue(op2())
        self.assertFalse(op3())

    def test_and_raises_nea(self):
        op = O.And(True)
        self.assertRaises(NotEnoughArguments, op)

    def test_or(self):
        op1 = O.Or(True, False)
        op2 = O.Or(True, True)
        op3 = O.Or(False, False)
        self.assertTrue(op1())
        self.assertTrue(op2())
        self.assertFalse(op3())

    def test_or_raises_nea(self):
        op = O.Or(True)
        self.assertRaises(NotEnoughArguments, op)



from .errors import NotEnoughArguments, ArgumentTypeError


class Func:
    """
    Базовая функция.
    """
    description = '__'  # описание функции
    args_description = []  #  описание аргументов
    group = 'str'  # группа, к которой относится функция (математическая, строковая и пр.)
    operation = None  # функция, которая будет выполняться

    """
    Признак того, что функция принимает аргументы одного типа.
    При None не производится проверка аргументов на соответствие этому типу.
    Если атрибут не None, то производится проверка сооветствия всех аргументов
    этому типу.
    """
    _arguments_type = None

    def __init__(self, *args):
        self._args = [arg for arg in args]  # переданные аргументы
        self._mandatory = 0
        for arg in self.__class__.args_description:
            if arg.is_mandatory:
                self._mandatory += 1

    @classmethod
    def get_func_info(cls):
        args = []
        for arg in cls.args_description:
            _ = ''
            if not arg.is_mandatory:
                _ = ' (необяз.)'
            args.append(f'{arg.name}{_}')

        name = cls.get_name()
        return {
            'group': cls.group,
            'name': name,
            'description': cls.get_description(),
            'args': args,
            'as_str': f'{name}({",".join(args)})',
        }

    @classmethod
    def get_name(cls):
        return cls.__name__.lower()

    @classmethod
    def get_description(cls):
        """
        Описание функции.
        """
        return cls.description

    @classmethod
    def get_group(cls):
        """
        Группа, к которой относится функция.
        """
        return cls.group

    @property
    def arguments_type(self):
        return self.__class__._arguments_type

    def _operation(self, *args):
        """
        Функция, которая будет выполняться,
        если operation класса is None.
        """
        pass

    def _get_operation(self):
        """
        Функция, которая будет выполняться.
        """
        cls_op = self.__class__.operation
        if cls_op is not None:
            return cls_op
        else:
            return self._operation

    def __call__(self):
        operation = self._get_operation()

        # проверка количества аргументов
        args_len = len(self._args)
        if args_len < self._mandatory:
            raise NotEnoughArguments(self.get_name(), self._mandatory)

        args = []
        for arg in self._args:
            if issubclass(type(arg), Func):
                args.append(arg())
            else:
                args.append(arg)

        # проверка типов аргументов
        if self.arguments_type is not None:
            arg_check = self._check_one_type_arguments
        else:
            arg_check = self._check_arguments
        arg_check(args)

        return operation(*args)

    def _check_arguments(self, arguments):
        args_desc = self.__class__.args_description
        for ind, value in enumerate(arguments):
            try:
                arg = args_desc[ind]
            except IndexError:
                break
            is_valid = arg.check_value(value)
            if not is_valid:
                type_ = value.__class__.__name__
                raise ArgumentTypeError(self.get_name(), type_)
        return True

    def _check_one_type_arguments(self, arguments):
        arg_type = self.arguments_type
        for arg in arguments:
            if not isinstance(arg, arg_type):
                type_ = arg.__class__.__name__
                raise ArgumentTypeError(self.get_name(), type_)
        return True


class MathFunc(Func):
    """
    Математическая функция.
    """
    group = 'math'


class StrFunc(Func):
    """
    Функция для работы со строками.
    """
    group = 'str'
    _arguments_type = str


class DateFunc(Func):
    """
    Функция для работы с датой и временем.
    """
    group = 'date'


class LogicFunc(Func):
    """
    Логическая функция.
    """
    group = 'logic'


class TypeconvFunc(Func):
    """
    Функция преобразования типа.
    """
    group = 'typeconv'


class DatastructFunc(Func):
    """
    Функция для работы со структурами данных.
    """
    group = 'datastruct'


class Operator(Func):
    """
    Функции-операторы.
    """
    group = 'operators'
    sign = None

    @property
    def sign(self):
        return self.__class__.sign


class DataframeFunc(Func):
    """
    Функции для работы с датафреймами.
    """
    group = 'dataframe'


class SqlFunc(Func):
    """
    Функции select-запросов.
    """
    group = 'sql'


class StatisticsFunc(Func):
    """
    Статистические функции.
    """
    group = 'statistics'


class ConnectionFunc(Func):
    """
    Функции подключения к источникам.
    """
    group = 'connection'


class NonMandatoryArg:
    """
    Необязательный аргумент функции.
    """
    mandatory = False  # обязательный ли аргумент

    def __init__(self, name, index, types=[]):
        self._name = name  # название
        self._index = index  # порядковый номер в списке аргументов
        self._types = types  # список допустимых типов аргумента

    @property
    def name(self):
        return self._name

    @property
    def is_mandatory(self):
        """
        Обязательный ли аргумент.
        """
        return self.__class__.mandatory

    def check_value(self, value):
        """
        Проверка значения на соответствие одному из допустимых типов.
        Возвращает True, если значение относится к одному из допустимых типов.
        При пустом списке типов (self._types) возвращается True, т.е. тип аргумента
        может быть любым.
        """
        type_is_ok = False
        if self._types == []:
            return True
        for type_ in self._types:
            if isinstance(value, type_):
                type_is_ok = True
                break
        return type_is_ok


class MandatoryArg(NonMandatoryArg):
    """
    Обязательный аргумент.
    """
    mandatory = True



from pandas import DataFrame
from unittest.mock import Mock

from noofa.tests.base import NoofaTest
from noofa.components.components import ComponentsSchema
from noofa.components.exceptions import SchemaComponentNotFound
from noofa.components.tables import ReportTable, ReportPivotTable
from noofa.components.figures import PlotlyLine


class TestComponentsSchema(NoofaTest):
    """
    Тестирование схемы компонентов.
    """
    @classmethod
    def setUpClass(cls):
        cls.evaluator = _mock_evaluator

    def setUp(self):
        self.schema = ComponentsSchema()

        self.table_conf = {
            'id': 'table1',
            'base': {
                'from': 'dataframe',
                'value': 'df1',
            },
            'title_text': 'Table',
            'using_evaluator': self.evaluator,
            'aliases': {'b': 'B'},
            'exclude': ['c'],
        }
        self.pivot_table_conf = {
            'id': 'pivottable1',
            'base': {
                'from': 'dataframe',
                'value': 'df1',
            },
            'title_text': 'Pivot table',
            'using_evaluator': self.evaluator,
            'pivot_conf': {
                'index': ['d'],
                'columns': [],
                'aggfunc': {'c': ['sum']},
                'values': {},
                'fill_value': None,
            },
        }
        self.figures_conf = _figures_conf

    def test_add_table(self):
        self.assertRaises(SchemaComponentNotFound, self.schema.get_table, 'table1')
        n1 = len(self.schema._tables.keys())
        self.schema.add_table(**self.table_conf)
        n2 = len(self.schema._tables.keys())
        self.assertEqual(n1 + 1, n2)

        table = self.schema.get_table('table1')
        self.assertIsInstance(table, ReportTable)

    def test_add_pivot_table(self):
        self.assertRaises(SchemaComponentNotFound, self.schema.get_table, 'pivottable1')
        n1 = len(self.schema._tables.keys())
        self.schema.add_pivot_table(**self.pivot_table_conf)
        n2 = len(self.schema._tables.keys())
        self.assertEqual(n1 + 1, n2)

        pt = self.schema.get_table('pivottable1')
        self.assertIsInstance(pt, ReportPivotTable)

    def test_add_figure(self):
        self.assertRaises(SchemaComponentNotFound, self.schema.get_figure, 'fig1')
        n1 = len(self.schema._figures.keys())
        self.schema.add_figure(**self.figures_conf['fig1'])
        n2 = len(self.schema._figures.keys())
        self.assertEqual(n1 + 1, n2)

        fig = self.schema.get_figure('fig1')
        self.assertIsInstance(fig, PlotlyLine)

    def test_table(self):
        self.schema.add_table(**self.table_conf)
        table = self.schema.get_table('table1')
        table.build()
        df = table.df
        self.assertIn('a', df)
        self.assertNotIn('b', df)
        self.assertIn('B', df)
        self.assertNotIn('c', df)

    def test_pivot_table(self):
        self.schema.add_pivot_table(**self.pivot_table_conf)
        pt = self.schema.get_table('pivottable1')
        pt.build()
        df = pt.pivot_df
        self.assertCountEqual(list(df['c']), [21, 10, 11])

    def test_build_lines_and_bars(self):
        for f, conf in self.figures_conf.items():
            self.schema.add_figure(**conf)
            fig = self.schema.get_figure(f)

            success = True
            try:
                fig.build()
            except:
                success = False

            with self.subTest(f):
                self.assertTrue(success)


_mock_evaluator = Mock()
_df = DataFrame({
    'a': [1, 2, 3, 4],
    'b': [5, 6, 7, 8],
    'c': [9, 10, 11, 12],
    'd': ['a', 'b', 'c', 'a'],
})
_mock_evaluator.evaluate.return_value = _df
_mock_evaluator.get_or_build_dataframe.return_value = _df

_pie_mock_evaluator = Mock()
_pie_mock_evaluator.get_or_build_dataframe.return_value = _df
_pie_mock_evaluator.evaluate.return_value = 1

_figures_conf = {
    'fig1': {
        'id': 'fig1',
        'base': {
            'from': 'dataframe',
            'value': {'dataframe': 'df1'},
            'x': 'a',
            'y': 'b',
        },
        'figure_type': 'line',
        'using_evaluator': _mock_evaluator,
    },
    'line1': {
        'id': 'line1',
        'base': {
            'from': 'grouped',
            'value': {'df_from': 'expression', 'dataframe': 'df1'},
            'x': 'a',
            'y': 'b',
            'line_group': 'd',
        },
        'figure_type': 'line',
        'using_evaluator': _mock_evaluator,
    },
    'line2': {
        'id': 'line2',
        'base': {
            'from': 'grouped',
            'value': {'df_from': 'dataframe', 'dataframe': 'df1'},
            'x': 'a',
            'y': 'b',
            'line_group': 'd',
        },
        'figure_type': 'line',
        'using_evaluator': _mock_evaluator,
    },
    'line3': {
        'id': 'line3',
        'base': {
            'from': 'list',
            'value': [
                {
                    'name': 'l1',
                    'x_from': 'expression',
                    'y_from': 'expression',
                    'x': 'df1',
                    'y': 'df1',
                },
            ],
        },
        'figure_type': 'line',
        'using_evaluator': _mock_evaluator,
    },
    'line4': {
        'id': 'line4',
        'base': {
            'from': 'list',
            'value': [
                {
                    'name': 'l1',
                    'x_from': 'column',
                    'y_from': 'column',
                    'x': {'df_from': 'expression', 'dataframe': 'df1', 'column': 'a'},
                    'y': {'df_from': 'dataframe', 'dataframe': 'df1', 'column': 'b'},
                },
            ],
        },
        'figure_type': 'line',
        'using_evaluator': _mock_evaluator,
    },
    'bar1': {
        'id': 'bar1',
        'base': {
            'from': 'dataframe',
            'value': {'dataframe': 'df1'},
            'x': 'a',
            'y': 'b',
            'barmode': 'stack',
        },
        'figure_type': 'bar',
        'using_evaluator': _mock_evaluator,
    },
    'bar2': {
        'id': 'bar2',
        'base': {
            'from': 'list',
            'value': [
                {
                    'name': 'b1',
                    'x_from': 'column',
                    'y_from': 'column',
                    'x': {'df_from': 'expression', 'dataframe': 'df1', 'column': 'a'},
                    'y': {'df_from': 'dataframe', 'dataframe': 'df1', 'column': 'b'},
                },
            ],
            'barmode': 'stack',
        },
        'figure_type': 'bar',
        'using_evaluator': _mock_evaluator,
    },
    'bar3': {
        'id': 'bar3',
        'base': {
            'from': 'agg',
            'value': {
                'dataframe': 'anything',
                'groupby': ['d'],
                'func': 'sum',
                'on': 'a',
            },
            'barmode': 'stack',
        },
        'figure_type': 'bar',
        'using_evaluator': _mock_evaluator,
    },
    'pie1': {
        'id': 'pie1',
        'base': {
            'from': 'dataframe',
            'value': {
                'dataframe': 'df1',
            },
            'names': 'd',
            'values': 'a',
        },
        'figure_type': 'pie',
        'using_evaluator': _mock_evaluator,
    },
    'pie2': {
        'id': 'pie2',
        'base': {
            'from': 'list',
            'value': [
                {'name': 'a1', 'value': 'some_expression'},
                {'name': 'a2', 'value': 'some_other_expression'},
            ],
        },
        'figure_type': 'pie',
        'using_evaluator': _pie_mock_evaluator,
    },
    'pie3': {
        'id': 'pie3',
        'base': {
            'from': 'agg',
            'value': {
                'dataframe': 'anything',
                'groupby': ['d'],
                'func': 'sum',
                'on': 'a',
            },
        },
        'figure_type': 'pie',
        'using_evaluator': _pie_mock_evaluator,
    },
}



from noofa.tests.base import NoofaInterpreterTest
from noofa.core.func.errors import (
    InterpreterContextError,
    ValueNotInContextError,
    RecursiveEvaluationError,
)


class TestContext(NoofaInterpreterTest):
    """
    Тестирование контекста интерпретатора.
    """
    @classmethod
    def setUpClass(cls):
        cls.set_up_interpreter()
        cls.nonex_value = 'non_existent_value'
        cls.ex_value = 'x'
        cls.value = 5
        cls.local_value = 4

    def tearDown(self):
        self.interpreter._context._values = {}
        self.interpreter._context.clear_local()
        self.interpreter._context.global_context = {}
        self.interpreter._context.switch_to_global()

    def test_var_not_found(self):
        self.assertRaises(
            InterpreterContextError,
            self.evaluate,
            self.nonex_value
        )

    def test_global_var(self):
        self.interpreter.add_to_global(
            self.ex_value,
            self.value
        )

        self.interpreter._context.switch_to_global()
        gv = self.evaluate(self.ex_value)
        self.assertEqual(gv, self.value)

    def test_local_var(self):
        self.interpreter.add_to_local(
            self.ex_value,
            self.local_value
        )

        self.interpreter._context.switch_to_local()
        lv = self.evaluate(self.ex_value)
        self.assertEqual(lv, self.local_value)

    def test_global_and_local(self):
        self.interpreter.add_to_global(
            self.ex_value,
            self.value
        )
        self.interpreter.add_to_local(
            self.ex_value,
            self.local_value
        )
        self.interpreter._context.switch_to_global()
        gv = self.evaluate(self.ex_value)
        self.interpreter._context.switch_to_local()
        lv = self.evaluate(self.ex_value)

        self.assertEqual(gv, self.value)
        self.assertEqual(lv, self.local_value)
        self.assertNotEqual(gv, lv)

    def test_value(self):
        self.interpreter.add_values(
            [{'name': self.ex_value, 'value': str(self.value)}],
            self.interpreter
        )
        self.assertEqual(
            self.evaluate(f'get_value("{self.ex_value}")'),
            self.value
        )
        self.assertEqual(
            self.interpreter.get_value(self.ex_value),
            self.value
        )

    def test_recursive_evaluation(self):
        expr = f'get_value("{self.ex_value}")'
        self.interpreter.add_values(
            [{'name': self.ex_value, 'value': expr}],
            self.interpreter
        )
        self.assertRaises(
            RecursiveEvaluationError,
            self.evaluate,
            expr
        )

    def test_non_existent_value(self):
        self.assertRaises(
            ValueNotInContextError,
            self.evaluate,
            f'get_value("{self.nonex_value}")'
        )



from pandas import DataFrame, Series
from datetime import datetime

from noofa.tests.base import NoofaFunctionTest
from noofa.core.func import typeconv as TC


class TestTypeConvFunctions(NoofaFunctionTest):
    """
    Тестирование функций преобразования типов.
    """
    @classmethod
    def setUpClass(cls):
        cls.add_cases_from_list(_CASES)


_dt = datetime.now()
_bad_args = [_dt, _dt.date(), _dt.time(), DataFrame(), Series((1,2))]

"""
Содержимое кортежей в списке _CASES:
(
    Функция,
    сабкейсы (кортеж из аргументов и ожидаемого результата),
    аргументы для проверки ошибки по типу аргумента,
    аргументы для проверки ошибки по кол-ву аргументов
)
"""
_CASES = [
    (TC.To_Str, (('1', '1'), (1, '1'), (1.1, '1.1')), _bad_args, ()),
    (TC.To_Float, (('1', 1.0), (1, 1.0), (1.1, 1.1)), _bad_args, ()),
    (TC.To_Int, (('1', 1), (1, 1), (1.1, 1)), _bad_args, ()),
]



class ExpressionParsingError(Exception):
    pass


class ExpressionSyntaxError(Exception):
    def __str__(self):
        return 'Ошибка синтаксиса'


class ExpressionEvaluationError(Exception):
    pass


class NotEnoughArguments(Exception):
    def __init__(self, fname, mandatory):
        self._fname = fname
        self._mandatory = mandatory

    def __str__(self):
        return f'Функции {self._fname} требуется минимум {self._mandatory} аргументов'


class ArgumentTypeError(Exception):
    def __init__(self, fname, arg_type):
        self._fname = fname
        self._arg_type = arg_type

    def __str__(self):
        return f'Функция {self._fname} не принимает аргументы типа {self._arg_type}'


class InterpreterContextError(Exception):
    def __init__(self, key):
        self.key = key

    def __str__(self):
        return f'В контексте отсутствует элемент {self.key}'


class ValueNotInContextError(Exception):
    def __init__(self, key):
        self.key = key

    def __str__(self):
        return f'В контексте отсутствует значение {self.key}'


class RecursiveEvaluationError(Exception):
    def __init__(self, key):
        self.key = key

    def __str__(self):
        return f'Рекурсивное вычисление при получении значения {self.key}'



import json
import redis
import sqlite3
import psycopg2
import pyodbc
import requests
import mysql.connector
from abc import ABC, abstractmethod
from random import choice

from .file_sources import FILE_SOURCES
from .statements import (
    _TEST_QUERIES,
    _TABLES_QUERIES,
    _COLUMNS_QUERIES,
    _FIELDS_QUERIES,
    _CONSTRAINTS_QUERIES,
)


_DEFAULT_MSSQL_DRIVER = '{ODBC Driver 18 for SQL Server}'
_FREE_TDS = '{FreeTDS}'


class DataQueryResult:
    def __init__(self, data, columns=[]):
        self.data = data
        self.columns = columns


class DataSource(ABC):
    """
    Абстрактный источник.
    """

    _is_sql = False
    source_type = ''

    @abstractmethod
    def get_connection(self):
        """
        Создание соединения.
        """
        pass

    @abstractmethod
    def get_data(self):
        """
        Получение данных.
        """
        pass

    @abstractmethod
    def test(self):
        """
        Проверка связи с источником.
        """
        pass

    @property
    def _test_query(self):
        return _TEST_QUERIES.get(self.source_type, None)

    @property
    def _tables_query(self):
        return _TABLES_QUERIES.get(self.source_type, None)

    @property
    def _columns_query(self):
        return _COLUMNS_QUERIES.get(self.source_type, None)

    @property
    def _fields_query(self):
        return _FIELDS_QUERIES.get(self.source_type, None)

    @property
    def _constraints_query(self):
        return _CONSTRAINTS_QUERIES.get(self.source_type, None)

    @property
    def is_sql(self):
        return self.__class__._is_sql

    @property
    def source_type(self):
        return self.__class__.source_type

    def _parse_conn_str(self):
        return _parse_conn_str(self._conn_str)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()


class DatabaseSource(DataSource):
    _is_sql = True
    _wildcard = '%s'
    enquote = False

    @property
    def wildcard(self):
        return self.__class__._wildcard

    @property
    def enquote(self):
        return self.__class__.enquote

    @abstractmethod
    def get_tables(self):
        """
        Получение списка таблиц в базе.
        Исключаются таблицы, содержащие системную информацию.
        """
        pass

    @abstractmethod
    def get_fields(self, table_name, **kwargs):
        """
        Получение списка полей в таблице.

        table_name - имя таблицы из бд.
        """
        pass

    def get_table(self, table_name):
        """
        Построение таблицы.

        table_name - имя таблицы из бд.
        """

        from .query import Table
        return Table(
            table_name,
            self.get_fields(table_name),
            enquote=self.enquote
        )

    def get_table_multiple(self, tables_list):
        from .query import Table

        assert isinstance(tables_list, list), 'tables_list должен быть списком'
        tables = {}
        fields = []

        if tables_list:
            _tables_list = ""
            for table_name in tables_list:
                _tables_list += f"'{table_name}', "
            _tables_list = _tables_list.rstrip(', ')

            with self.connection.cursor() as cursor:
                q = "SELECT column_name, data_type, table_name FROM information_schema.columns "
                q += "WHERE table_name in (%s) ORDER BY table_name" % _tables_list
                cursor.execute(q)
                fields = cursor.fetchall()

        for field_name, data_type, table_name in fields:
            if not table_name in tables:
                tables[table_name] = []
            table = tables[table_name]
            table.append(field_name)

        tables = {
            table_name: Table(table_name, table_fields, enquote=self.enquote)
            for table_name, table_fields in tables.items()
        }

        return tables

    def get_db_structure(self):
        with self.connection.cursor() as cursor:
            cursor.execute(self._columns_query, (self._dbname, ))
            columns_res = cursor.fetchall()
            cursor.execute(self._constraints_query)
            constraints_res = cursor.fetchall()

        return _db_struct(columns=columns_res,
                constraints=constraints_res)


class PostgresSource(DatabaseSource):
    """
    Источник postgres.
    """

    source_type = 'postgres'
    enquote = True

    def __init__(self, host=None, port=5432, dbname=None,
        user=None, password=None, conn_str=None, **kwargs):
        self._dbname = dbname
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._conn_str = conn_str
        self.connection = None

    def get_tables(self):
        tables = []
        with self.connection.cursor() as cursor:
            q = self._tables_query
            cursor.execute(q)
            res = cursor.fetchall()
        tables = [r[0] for r in res]
        return tables

    def get_connection(self):
        if self._conn_str:
            conn_dict = self._parse_conn_str()
            conn_dict['dbname'] = conn_dict.pop('database')
            conn_dict['connect_timeout'] = 3
            return psycopg2.connect(**conn_dict)
        conn = psycopg2.connect(
            host=self._host,
            port=self._port,
            dbname=self._dbname,
            user=self._user,
            password=self._password,
            connect_timeout=3,
        )
        return conn

    def open(self):
        print('open')
        if self.connection is not None:
            self.connection.close()
        self.connection = self.get_connection()

    def close(self):
        print('close')
        self.connection.commit()
        self.connection.close()
        self.connection = None

    def test(self):
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(self._test_query)
        except:
            return False
        return True

    def get_data(self, query=None, **kwargs):
        q, params = query.str_and_params()
        fields = query.requested
        data = []

        with self.connection.cursor() as cursor:
            if params:
                cursor.execute(q, params)
            else:
                cursor.execute(q)

            _data = cursor.fetchall()
            desc = [f.replace('"', "") for f in fields]
            for d in _data:
                datapiece = {}
                for n, i in enumerate(d):
                    datapiece[desc[n]] = i
                data.append(datapiece)
        return DataQueryResult(data, desc)

    def get_fields(self, table_name):
        fields = []
        with self.connection.cursor() as cursor:
            q = self._fields_query
            cursor.execute(q, (table_name, ))
            res = cursor.fetchall()
            fields = [column[0] for column in res]
        return fields


class MySqlSource(DatabaseSource):
    """
    Источник mysql.
    """

    source_type = 'mysql'

    def __init__(self, host=None, port=3306, dbname=None,
        user=None, password=None, conn_str=None, **kwargs):
        self._dbname = dbname
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._conn_str = conn_str
        self.connection = None

    def get_tables(self):
        tables = []
        with self.connection.cursor() as cursor:
            q = self._tables_query
            cursor.execute(q, (self._dbname, ))
            res = cursor.fetchall()
        tables = [r[0] for r in res]
        return tables

    def get_connection(self):
        if self._conn_str:
            conn_dict = self._parse_conn_str()
            conn_dict['connect_timeout'] = 3
            return mysql.connector.connect(**conn_dict)
        conn = mysql.connector.connect(
            host=self._host,
            port=self._port,
            database=self._dbname,
            user=self._user,
            password=self._password,
            connect_timeout=3,
        )
        return conn

    def open(self):
        if self.connection is not None:
            self.connection.close()
        self.connection = self.get_connection()

    def close(self):
        self.connection.close()
        self.connection = None

    def test(self):
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(self._test_query)
        except:
            return False
        return True

    def get_data(self, query=None, **kwargs):
        q, params = query.str_and_params()
        fields = query.requested
        data = []
        with self.connection.cursor() as cursor:
            if params:
                cursor.execute(q, params)
            else:
                cursor.execute(q)

            _data = cursor.fetchall()
            desc = fields
            for d in _data:
                datapiece = {}
                for n, i in enumerate(d):
                    datapiece[desc[n]] = i
                data.append(datapiece)
        return DataQueryResult(data, desc)

    def get_fields(self, table_name):
        fields = []
        with self.connection.cursor() as cursor:
            q = self._fields_query
            cursor.execute(q, (table_name, self._dbname, ))
            res = cursor.fetchall()
            fields = [column[0] for column in res]
        return fields


class MSSqlSource(DatabaseSource):
    """
    Источник mssql.
    """

    source_type = 'mssql'
    enquote = True
    _wildcard = '?'

    def __init__(self, ad=False, host=None, port=1433,
        dbname=None, user=None, password=None, conn_str=None, **kwargs):
        self._ad = ad
        self._driver = _FREE_TDS if ad == True else _DEFAULT_MSSQL_DRIVER
        self._host = host
        self._port = port
        self._dbname = dbname
        self._user = user
        self._password = password
        self._conn_str = conn_str
        self.connection = None

    def get_tables(self):
        tables = []
        q = self._tables_query
        with self.connection.cursor() as cursor:
            cursor.execute(q)
            res = cursor.fetchall()
        tables = [r[0] for r in res]
        return tables

    def get_connection(self):
        conn_str = self._make_conn_str()
        conn = pyodbc.connect(conn_str, timeout=3)
        return conn

    def open(self):
        if self.connection is not None:
            self.connection.close()
        self.connection = self.get_connection()

    def close(self):
        self.connection.close()
        self.connection = None

    def test(self):
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(self._test_query)
        except:
            return False
        return True

    def get_data(self, query=None, **kwargs):
        q, params = query.str_and_params()
        fields = query.requested
        data = []

        with self.connection.cursor() as cursor:
            if params:
                cursor.execute(q, params)
            else:
                cursor.execute(q)

            _data = cursor.fetchall()
            desc = [f.replace('"', "") for f in fields]
            for d in _data:
                datapiece = {}
                for n, i in enumerate(d):
                    datapiece[desc[n]] = i
                data.append(datapiece)
        return DataQueryResult(data, desc)

    def get_fields(self, table_name):
        fields = []
        with self.connection.cursor() as cursor:
            q = self._fields_query
            cursor.execute(q, (table_name, ))
            res = cursor.fetchall()
            fields = [column[0] for column in res]
        return fields

    def _make_conn_str(self):
        if self._conn_str:
            conn_dict = self._parse_conn_str()
            ad = conn_dict.get('ad', 'false')
            driver = _FREE_TDS if ad == 'true' else _DEFAULT_MSSQL_DRIVER
            host, port = conn_dict['host'], conn_dict.get('port', self._port)
            server = f'{host},{port}'
            db = conn_dict['database']
            user, password = conn_dict['user'], conn_dict['password']
        else:
            driver = self._driver
            server = f'{self._host},{self._port}'
            db = self._dbname
            user, password = self._user, self._password

        conn_str = f'DRIVER={driver};SERVER={server};DATABASE={db};'
        conn_str += f'UID={user};PWD={password};ENCRYPT=optional;'

        return conn_str

    def get_db_structure(self):
        with self.connection.cursor() as cursor:
            cursor.execute(self._columns_query)
            columns_res = cursor.fetchall()
            cursor.execute(self._constraints_query)
            constraints_res = cursor.fetchall()

        return _db_struct(columns=columns_res,
                constraints=constraints_res)


class SqliteSource(DatabaseSource):
    """
    Источник sqlite. Используется для демонстрационной конфигурации -
    noofa.examples. Не планируется использовать его как один из доступных
    источников.
    """

    source_type = 'sqlite'
    _wildcard = '?'

    def __init__(self, **kwargs):
        self._db_file = kwargs.get('dbname')
        self._conn_str = kwargs.get('conn_str', None)
        self.connection = None

    def get_tables(self):
        tables = []
        cursor = self.connection.cursor()
        q = self._tables_query
        cursor.execute(q)
        res = cursor.fetchall()
        tables = [r[0] for r in res]
        return tables

    def get_connection(self, **kwargs):
        if self._conn_str:
            conn_dict = self._parse_conn_str()
            return sqlite3.connect(conn_dict['database'])
        conn = sqlite3.connect(self._db_file)
        return conn

    def open(self):
        if self.connection is not None:
            self.connection.close()
        self.connection = self.get_connection()

    def close(self):
        self.connection.close()
        self.connection = None

    def test(self):
        try:
            cursor = self.connection.cursor()
            cursor.execute(self._test_query)
            cursor.close()
        except:
            return False
        return True

    def get_data(self, **kwargs):
        query = kwargs['query']
        q, params = query.str_and_params()
        fields = query.requested
        data = []

        cursor = self.connection.cursor()
        if params:
            cursor.execute(q, params)
        else:
            cursor.execute(q)

        _data = cursor.fetchall()
        desc = [f.replace('"', "") for f in fields]
        for d in _data:
            datapiece = {}
            for n, i in enumerate(d):
                datapiece[desc[n]] = i
            data.append(datapiece)
        return DataQueryResult(data, desc)

    def get_fields(self, table_name, **kwargs):
        fields = []
        cursor = self.connection.cursor()
        q = self._fields_query % table_name
        cursor.execute(q)
        fields = [description[0] for description in cursor.description]
        return fields

    def get_table_multiple(self, tables_list):
        from .query import Table

        assert isinstance(tables_list, list), 'tables_list должен быть списком'
        tables = {
            table_name: Table(table_name, self.get_fields(table_name))
            for table_name in tables_list
        }

        return tables


class RedisSource(DataSource):
    """
    Источник redis.
    """

    source_type = 'redis'

    def __init__(self, host='localhost', port=6379, db=0, user=None,
        password=None, source='', conn_str=None, **kwargs):
        self._host = host
        self._port = port
        self._db = db
        self._username = user
        self._password = password
        self.source = source
        self._conn_str = conn_str
        self.connection = None

    def get_connection(self):
        if self._conn_str:
            conn_dict = self._parse_conn_str()
            conn_dict['db'] = conn_dict.pop('database')
            return redis.Redis(**conn_dict)
        conn = redis.Redis(
            host=self._host,
            port=self._port,
            db=self._db,
            username=self._username,
            password=self._password,
            socket_timeout=3,
        )
        return conn

    def open(self):
        if self.connection is not None:
            self.connection.close()
        self.connection = self.get_connection()

    def close(self):
        self.connection.close()
        self.connection = None

    def test(self):
        try:
            result = self.connection.ping()
        except:
            return False
        return result

    def get_data(self):
        data, columns = [], []
        with self.connection as conn:
            _data = conn.hgetall(self.source)

            for k, d in _data.items():
                data.append(json.loads(d))
            if data:
                columns = [key for key in data[0].keys()]
        return DataQueryResult(data, columns)

    def get_tables(self):
        tables = []
        try:
            with self.connection as conn:
                for h in conn.scan_iter(_type='hash'):
                    tables.append(h.decode())
        except:
            pass
        return tables

    def get_fields(self):
        fields = []
        with self.connection as conn:
            if conn.hlen(self.source):
                rand_key = choice(conn.hkeys(self.source))
                rand_value = conn.hget(self.source, rand_key)
                rand_value = json.loads(rand_value)
                fields = list(rand_value.keys())
        return fields


class JsonSource(DataSource):
    """
    Источник из json-ответа.
    Использование среди прочих источников под вопросом.
    """

    source_type = 'json'

    def __init__(self, url, **kwargs):
        self._url = url
        self._params = kwargs.get('params', {})
        self._headers = kwargs.get('headers', {})
        self._auth = kwargs.get('auth', ())
        self.connection = None

    def get_connection(self, **kwargs):
        conn = requests.Session()
        return conn

    def open(self):
        if self.connection is not None:
            self.connection.close()
        self.connection = self.get_connection()

    def close(self):
        self.connection.close()
        self.connection = None

    def get_data(self, **kwargs):
        data = {}
        with self.connection as conn:
            resp = conn.get(
                self._url,
                headers=self._headers,
                auth=self._auth,
                params=self._params
            )
            data = resp.json()
        return DataQueryResult(data)

    def get_fields(self, **kwargs):
        return []


SOURCES_DICT = {
    'postgres': PostgresSource,
    'mysql': MySqlSource,
    'mssql': MSSqlSource,
    'redis': RedisSource,
    'json': JsonSource,
    'sqlite': SqliteSource,
}
SOURCES_DICT.update(FILE_SOURCES)


def get_source_class(type_):
    return SOURCES_DICT.get(type_, None)


def _parse_conn_str(conn_str):
    """
    Парсинг строки соединения.
    """
    if conn_str.endswith(';'):
        conn_str = conn_str[:-1]
    _ = ['host', 'port', 'database', 'user', 'password', 'ad']
    conn_dict = {}
    conn_str = conn_str.split(';')
    for s in conn_str:
        spl = s.split('=')
        arg, val = spl[0], spl[1]
        if arg in _:
            conn_dict[arg] = val
    return conn_dict


def _db_struct(columns=[], constraints=[]):
    db = {
        'tables': {},
        'relations': [],
    }

    for c in columns:
        table_name, column_name, data_type = c
        table = db['tables'].setdefault(table_name, {
            'name': table_name,
            'columns': [],
            'pk': [],
            'fk': [],
            'fk_cols': [],
            'relates': 0,
            'related_to': 0,
            'related_by': [],
        })
        table['columns'].append({
            'name': column_name,
            'type': data_type,
        })

    for con in constraints:
        (table_name, constr_type, column_name,
            referenced_table, referenced_column) = con
        if constr_type.lower() == 'primary key':
            pk_list = db['tables'][table_name]['pk']
            if not column_name in pk_list:
                pk_list.append(column_name)
        elif constr_type.lower() == 'foreign key':
            db['tables'][table_name]['relates'] += 1
            db['tables'][referenced_table]['related_to'] += 1
            db['tables'][table_name]['fk'].append(referenced_table)
            db['tables'][table_name]['fk_cols'].append(column_name)
            db['tables'][referenced_table]['related_by'].append(table_name)
            db['relations'].append({
                'referencing_table': table_name,
                'referencing_column': column_name,
                'referenced_table': referenced_table,
                'referenced_column': referenced_column,
            })
    return db



from noofa.tests.base import NoofaSuite, NoofaRunner
from noofa.tests.builders.builder import TestBuilder


def builders_suite():
    suite = NoofaSuite()
    suite.add(TestBuilder)
    return suite


def run_tests():
    suite = builders_suite()
    runner = NoofaRunner()
    runner.run(suite)



from . import query


_JOINS = {
    'inner': query.Join,
    'left': query.LeftJoin,
    'right': query.RightJoin,
}

_FILTERS = {
    '==': query.EqFilter,
    '!=': query.NeqFilter,
    '>': query.GtFilter,
    '>=': query.GeFilter,
    '<': query.LtFilter,
    '<=': query.LeFilter,
    'contains': query.ContainsFilter,
    'startswith': query.StartsWithFilter,
    'endswith': query.EndsWithFilter,
    'in': query.InFilter,
    'not in': query.NotInFilter,
}


def collect_tables(query_as_json):
    subqueries = _find_subqueries(query_as_json)
    tables = _find_tables([query_as_json] + subqueries)
    return list(set(tables))

def _find_subqueries(query):
    """
    Поиск подзапросов в запросе в виде словаря.
    """
    res = []

    try:
        query_filters = query.get('filters', [])
    except AttributeError:
        query_filters = query
    for filter_ in query_filters:
        is_q = filter_['is_complex']
        if is_q:
            res += _find_subqueries(filter_)
        else:
            val = filter_['value']
            try:
                is_subquery = val['is_subquery']
            except:
                continue
            else:
                if is_subquery:
                    res.append(val)
                filters = val.get('filters', [])
                res += _find_subqueries(filters)
    return res

def _find_tables(queries):
    """
    Получение списка таблиц по списку запросов в виде словаря.
    """
    tables = []
    for s in queries:
        _tables = s.get('tables', [])
        if _tables:
            tables += _tables
    return tables


class Qbuilder:
    """
    Конструктор запросов.
    """

    # В конструктов передаются:
    # словарь используемых в запросе таблиц вида {название таблицы: объект Table}
    # и запрос в виде словаря
    def __init__(self, tables, query, param_placeholder='%s', mssql=False):
        self._source = None  # источник данных
        self._base = query['base']
        self._tables = tables  # словарь таблиц
        self._base_table = self._tables[self._base]  # базовая таблица
        self._joins_list = query.get('joins', [])  # список соединений в запросе
        self._filters_list = query.get('filters', [])  # список фильтров в запросе
        self._values_list = query.get('values', [])  # список запрашиваемых полей
        self._limit = query.get('limit', None)  #  limit в запросе
        self._orderings = query.get('order_by', [])  # упорядочивание
        self._is_mssql = mssql

        try:
            int(self._limit)
        except:
            self._limit = None
        self._param_placeholder = param_placeholder

    def parse_joins(self):
        """
        Построение соединений таблиц.
        Возвращается список объектов Join либо производных от Join классов.
        """
        joins = []

        for join in self._joins_list:
            l, r, j = join['l'], join['r'], join['j']
            on_l, on_r = join['on']['l'], join['on']['r']
            J = _JOINS.get(j, 'inner')
            l, r = self._tables[l], self._tables[r]
            l.has_field(on_l), r.has_field(on_r)
            joins.append(J(l, r, (on_l, on_r)))

        return joins

    def _parse_filters(self, f):
        """
        Построение объекта составного фильтра Q.
        """
        if f['is_complex'] == False:
            # если фильтр простой 
            # создать простой фильтр
            _filter = _FILTERS.get(f['op'], None)
            if _filter:
                val = f['value']
                # парсить подзапрос, если он является значением фильтра
                try:
                    is_subquery = val['is_subquery']
                except:
                    pass
                else:
                    if is_subquery:
                        qb = Qbuilder(self._tables, val)
                        val = qb.parse_query()

                table_name = f['table']
                field_name = f['field']
                table = self._tables[table_name]
                table.has_field(field_name)
                if table.enquote:
                    field_name = f'"{table_name}"."{field_name}"'
                else:
                    field_name = f'{table_name}.{field_name}'
                _filter = _filter(field_name, val, param_placeholder=self._param_placeholder)
                return query.Q(_filter)
        else:
            op = f['op'].lower()
            res = []
            for i in f['filters']:
                res.append(self._parse_filters(i))

            q = query.Q()
            for i in res:
                if op == 'or':
                    q |= query.Q(i)
                else:
                    q &= query.Q(i)

            return q

    def parse_filters(self):
        """
        Построение фильтров.
        Возвращается список составных фильтров Q.
        """
        res = []

        for f in self._filters_list:
            res.append(self._parse_filters(f))

        return res

    def parse_query(self):
        """
        Построение запроса.
        Возвращается объект запроса SelectQuery.
        """
        joins, filters = self.parse_joins(), self.parse_filters()
        values, orderings = self._parse_values(), self._parse_orderings()
        q = self._base_table.select().join(*joins).where(*filters).values(*values)
        q = q.order_by(*orderings)

        if self._limit is not None:
            l = self._limit
            q.limit(l) if not self._is_mssql else q.mssql_limit(l)
        return q

    def _parse_values(self):
        result = []
        values_list = self._values_list
        for v in values_list:
            table = self._tables[v['table']]
            field_name = v['field']
            try:
                alias = v['alias']
            except KeyError:
                alias = None
            table.has_field(field_name)
            if table.enquote:
                value = f'"{table._name}"."{field_name}"'
            else:
                value = f'{table._name}.{field_name}'
            result.append((value, alias))
        return result

    def _parse_orderings(self):
        result = []
        for ordering in self._orderings:
            table = self._tables[ordering['table']]
            type_ = ordering['type'].upper()
            fields = ordering['fields']
            values = []
            for field_name in fields:
                table.has_field(field_name)
                if table.enquote:
                    value = f'"{table._name}"."{field_name}"'
                else:
                    value = f'{table._name}.{field_name}'
                if type_ != 'DESC':
                    type_ = 'ASC'
                values.append(value)
            result.append((values, type_))
        return result


def collect_query_filters():
    return list(_FILTERS.keys())



class RecursiveDataframeBuildError(Exception):
    def __init__(self, dataframe_id):
        self.dataframe_id = dataframe_id

    def __str__(self):
        return f'Перекрёстная ссылка при построении датафрейма {self.dataframe_id}'


from pandas import DataFrame, Series

from noofa.tests.base import NoofaTest
from noofa.core.func import datastruct as DS
from noofa.core.func.errors import (
    NotEnoughArguments,
    ArgumentTypeError,
)


class TestDatastructFunctions(NoofaTest):
    """
    Тестирование функций работы со списками.
    """
    @classmethod
    def setUpClass(cls):
        cls.list = (1, 2, 3)
        cls.series = Series(cls.list)
        cls.df = DataFrame({'a': cls.list})

    def test_empty_list(self):
        f = DS.List()
        self.assertEqual(f(), [])

    def test_list(self):
        f = DS.List(1, 2, 3)
        self.assertEqual(f(), [1, 2, 3])

    def test_series_to_list(self):
        f = DS.ToList(self.series)
        self.assertEqual(f(), [1, 2, 3])

    def test_df_to_list(self):
        f = DS.ToList(self.df)
        self.assertEqual(f(), [1, 2, 3])

    def test_to_list_raises_nea(self):
        f = DS.ToList()
        self.assertRaises(NotEnoughArguments, f)

    def test_to_list_raises_ate(self):
        for arg in ['1', 1, 2.1, False]:
            f = DS.ToList(arg)
            with self.subTest(arg):
                self.assertRaises(ArgumentTypeError, f)



"""
Интерпретатор функций.
"""
from .parser import parse
from .context import Context, _context_funcs, GetFromContext, GetSlice
from ._functions import FUNCTIONS_DICT as _functions_dict
from ._functions import OPERATORS_DICT as _operators_dict
from .operators import _OPERATORS_PRIORITY
from .errors import ExpressionSyntaxError


_OPERATORS = _OPERATORS_PRIORITY.keys()


class Interpreter:
    """
    Интерпретатор для выполнения функций и вычислений.
    """
    def __init__(self, **context):
        self._functions_dict = {**_functions_dict}
        self._functions_dict.update(_context_funcs)
        self._operators_dict = _operators_dict
        self._context = Context(**context)
        self._connections = {}

    def add_values(self, values_list, evaluator):
        self._context.add_values(values_list, evaluator)

    def get_value(self, name):
        return self._context.get_value(name)

    def update_value(self, name, value):
        self._context.update_value(name, value)

    def evaluate(self, expression):
        """
        Получение результата формулы.
        """
        if not expression.endswith(';'):
            expression += ';'
        stree = parse(expression)
        stree = self._normalize_operators(stree[0])
        is_valid = self._check_syntax(stree)
        if not is_valid:
            raise ExpressionSyntaxError
        func = self._eval(stree)
        result = func()
        return result

    def apply(self, df, expr):
        """
        Применение выражения к строкам датафрейма.
        Здесь исользуется локальный контекст интерпретатора,
        при этом в выражении будут доступны переменные row - строка датафрейма в виде словаря,
        и idx - порядковый номер строки. Нумерация строк - с 0.
        """
        result = []
        rows = df.to_dict(orient='records')
        self._context.switch_to_local()
        key = 'row'
        for idx, row in enumerate(rows):
            self.add_to_local(key, row)
            self.add_to_local('idx', idx)
            result.append(self.evaluate(expr))
        if result:
            self._context.remove(key)
            self._context.remove('idx')
        self._context.switch_to_global()
        return result

    def add_to_global(self, key, value):
        """
        Добавление ключа-значения в глобальный контекст.
        """
        self._context.add_to_global(key, value)

    def add_to_local(self, key, value):
        """
        Добавление ключа-значения в локальный контекст.
        """
        self._context.add_to_local(key, value)

    def _check_syntax(self, stree):
        """
        Проверка синтаксиса выражения/части выражения.
        При наличии ошибки возвращает False.
        """
        if stree is None:
            return True
        type_ = stree['type']
        if type_ == 'symbol':
            return True
        if type_ == 'number':
            if stree['value'].count('.') > 1:
                return False
        if type_ == 'operator':
            sign = stree['value']
            left, right = stree['left'], stree['right']
            for o in (left, right):
                if not self._check_syntax(o):
                    return False
            if left is None:
                if not sign in ('+', '-'):
                    return False
            if right is None:
                return False
        if type_ == 'context':
            args = stree['args']
            for arg in args:
                if not self._check_syntax(arg):
                    return False
            var_ = stree['var']
            if var_ is None:
                return False
            if len(args) == 0:
                return False
        if type_ == 'call':
            args = stree['args']
            for arg in args:
                if not self._check_syntax(arg):
                    return False
            fname = stree['function']
            args_len = len(args)
            if fname is None:
                if not args_len or args_len > 1:
                    return False
        return True

    def _eval(self, stree):
        """
        Интерпретация синтаксического дерева с возвращением результата.
        """
        if stree is None:
            return None
        type_ = stree['type']
        if type_ == 'symbol':
            _func = self._get_function(GetFromContext.get_name())
            return _func(self._context, stree['value'])
        if type_ == 'string':
            func = self._get_function('to_str')
            return func(stree['value'])
        if type_ == 'number':
            value = stree['value']
            if '.' in value:
                _func = 'to_float'
            else:
                _func = 'to_int'
            func = self._get_function(_func)
            return func(value)
        if type_ == 'operator':
            sign = stree['value']
            left, right = self._eval(stree['left']), self._eval(stree['right'])
            func = self._get_operator(sign)
            return func(left, right)
        if type_ == 'context':
            slice_ = []
            for arg in stree['args']:
                r = self._eval(arg)
                slice_.append(r())
            var_ = stree['var']['value']
            _func = self._get_function(GetSlice.get_name())
            var_ = self._context.get(var_)
            return _func(var_, slice_)
        if type_ == 'call':
            args = []
            for arg in stree['args']:
                r = self._eval(arg)
                args.append(r)
            _func = stree['function']
            if _func is None:
                return args[0]
            else:
                fname = _func['value']
                func = self._get_function(fname)
                if func.group == 'context':
                    args.insert(0, self._context)
                #if _func['value'] == 'connection':
                #    return func(self._connections, *args)
            return func(*args)

    def _get_function(self, name):
        return self._functions_dict.get(name, None)

    def _get_operator(self, sign):
        return self._operators_dict.get(sign, None)

    def _normalize_operators(self, expr):
        """
        Упорядочивание операторов в выражении по приоритету операций.
        """
        type_ = expr['type']
        if type_ == 'operator':
            _unpacked = self._unpack_operator(expr)
            expr = self._sort_operators(_unpacked)
        elif type_ == 'call':
            args = []
            for arg in expr['args']:
                args.append(self._normalize_operators(arg))
            expr['args'] = args
        return expr

    def _sort_operators(self, unpacked_operator):
        """
        Сортировка операторов по приоритету операций.
        """
        def upd(unp, n, operator):
            left, right = unp[n-1], unp[n+1]
            op = {'type': 'operator', 'value': operator, 'left': left, 'right': right}
            l, r = unp[:n-1], unp[n+2:]
            l.append(op)
            return op, l + r

        def get_prioritized(priority):
            return [s for s, p in _OPERATORS_PRIORITY.items() if p == priority]

        res = []
        current_priority = 1
        prioritized_operators = get_prioritized(current_priority)
        # перебор распакованного оператора производится до тех пор,
        # пока он не будет приведён к одному словарю
        while len(unpacked_operator) != 1:
            for n, i in enumerate(unpacked_operator):
                type_ = type(i)
                if type_ is str:
                    priority = _OPERATORS_PRIORITY[i]
                    if priority == current_priority:
                        res, unpacked_operator = upd(unpacked_operator, n, i)
                        break
                    else:
                        keep_looking = False
                        for p in prioritized_operators:
                            if p in unpacked_operator:
                                keep_looking = True
                                break
                        if not keep_looking:
                            #res, unpacked_operator = upd(unpacked_operator, n, i)
                            current_priority += 1
                            prioritized_operators = get_prioritized(current_priority)
                            break
        return res

    def _unpack_operator(self, op):
        """
        Распаковка аргументов оператора.
        Используется для дальнейшего расположения
        арифмет. операторов в порядке приоритета операций.

        На выходе имеем список вида [токен1, оператор1, ..., операторN-1, токенN]
        """
        res = [op['value']]
        for i in ('left', 'right'):
            o = op[i]
            if o is None:
                if i == 'left':
                    res.insert(0, o)
                else:
                    res.append(o)
                continue
            if o['type'] == 'operator':
                unp = self._unpack_operator(o)
                if i == 'left':
                    for i in unp[::-1]:
                        res.insert(0, i)
                else:
                    for i in unp:
                        res.append(i)
            else:
                if i == 'left':
                    res.insert(0, o)
                else:
                    res.append(o)
        return res



from .exceptions import SchemaComponentNotFound
from .tables import ReportTable, ReportPivotTable
from .figures import FIGURES


class ComponentsSchema:
    """
    Схема компонентов отчёта.
    Содержит словари с таблицами и графиками.
    """
    def __init__(self, dataframes=None, tables=None, figures=None, **kwargs):
        self._dataframes = dataframes or {}
        self._tables = tables or {}
        self._figures = figures or {}

    def add_table(self, id=None, base=None, using_evaluator=None, **options):
        build_from, base_value = base.pop('from'), base.pop('value')

        self._tables[id] = ReportTable(
            id=id,
            build_from=build_from,
            base_value=base_value,
            evaluator=using_evaluator,
            **options,
        )

    def add_pivot_table(self, id=None, base=None, pivot_conf=None,
        using_evaluator=None, **options):
        build_from, base_value = base.pop('from'), base.pop('value')

        self._tables[id] = ReportPivotTable(
            id=id,
            build_from=build_from,
            base_value=base_value,
            pivot_conf=pivot_conf,
            evaluator=using_evaluator,
            **options,
        )

    def add_figure(self, id=None, base=None, using_evaluator=None, engine='plotly',
        figure_type=None, **options):
        build_from, base_value = base['from'], base['value']
        options.update({
            'x': base.get('x', ''),
            'y': base.get('y', ''),
            'values': base.get('values', ''),
            'names': base.get('names', ''),
            'line_group': base.get('line_group', ''),
            'barmode': base.get('barmode', ''),
            'labels': base.get('labels', ''),
        })

        base_cls = FIGURES[engine][figure_type]
        self._figures[id] = base_cls(
            id=id,
            build_from=build_from,
            base_value=base_value,
            evaluator=using_evaluator,
            **options,
        )

    def get_component(self, component_id):
        try:
            return self.get_table(component_id)
        except SchemaComponentNotFound:
            try:
                return self.get_figure(component_id)
            except:
                raise SchemaComponentNotFound(component_id, '_')

    def get_table(self, table_id):
        table = self._tables.get(table_id, None)
        if table is None:
            raise SchemaComponentNotFound(table_id, 'table')
        return table

    def get_figure(self, figure_id):
        figure = self._figures.get(figure_id, None)
        if figure is None:
            raise SchemaComponentNotFound(figure_id, 'figure')
        return figure



from pandas import DataFrame
from noofa.tests.base import NoofaTest
from noofa.core.dataframes import filters


class TestPandaFilters(NoofaTest):
    """
    Тестирование фильтров датафреймов.
    """
    @classmethod
    def setUpClass(cls):
        cls.df = DataFrame([
            {'a': 1, 'b': 'Hello', 'c': 3.14},
            {'a': 2, 'b': 'hell', 'c': 3.14},
            {'a': 2, 'b': 'ello', 'c': 3.14},
            {'a': 3, 'b': 'ohell', 'c': 3.14},
            {'a': 3, 'b': 'hellno', 'c': 3.14},
            {'a': 3, 'b': 'holle', 'c': 3.14},
        ])

        cls.filter1 =     {
            'col_name': 'a',
            'op': '>',
            'value': 1,
            'is_q': False,
        }
        cls.filter2 =     {
            'col_name': 'a',
            'op': '<',
            'value': 3,
            'is_q': False,
        }

    def test_eq(self):
        pf = filters.PandaEq('a', 1)
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 1)

    def test_neq(self):
        pf = filters.PandaNeq('a', 1)
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 5)

    def test_gt(self):
        pf = filters.PandaGt('a', 1)
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 5)

    def test_gte(self):
        pf = filters.PandaGte('a', 1)
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 6)

    def test_lt(self):
        pf = filters.PandaLt('a', 3)
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 3)

    def test_lte(self):
        pf = filters.PandaLte('a', 3)
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 6)

    def test_contains(self):
        pf = filters.PandaContains('b', 'el')
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 5)

    def test_startswith(self):
        pf = filters.PandaStartsWith('b', 'he')
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 3)

    def test_endswith(self):
        pf = filters.PandaEndsWith('b', 'lo')
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 2)

    def test_in(self):
        pf = filters.PandaIn('b', ['holle', 'hell'])
        pf.df = self.df
        filtered = self.df[pf.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 2)

    def test_empty_panda_q(self):
        pq = filters.PandaQ()
        self.assertTrue(pq.is_empty)

    def test_simple_panda_q(self):
        pf = filters.PandaIn('b', ['holle', 'hell'])
        pf.df = self.df
        pq = filters.PandaQ(pf)
        filtered = self.df[pq.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 2)

    def test_panda_q_and(self):
        pf1 = filters.PandaGt('a', 1)
        pf1.df = self.df
        pf2 = filters.PandaLt('a', 3)
        pf2.df = self.df

        pq = filters.PandaQ(pf1) & filters.PandaQ(pf2)
        filtered = self.df[pq.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 2)

    def test_panda_q_or(self):
        pf1 = filters.PandaEq('a', 1)
        pf1.df = self.df
        pf2 = filters.PandaEq('a', 3)
        pf2.df = self.df

        pq = filters.PandaQ(pf1) | filters.PandaQ(pf2)
        filtered = self.df[pq.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 4)

    def test_parse_filter(self):
        f = filters._parse_filter(self.df, self.filter1)
        self.assertIsInstance(f, filters.PandaQ)
        filtered = self.df[f.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 5)

    def test_parse_complex_filter(self):
        cmplx = {
            'is_q': True,
            'op': 'and',
            'filters': [
                self.filter1,
                self.filter2
            ]
        }
        f = filters._parse_filter(self.df, cmplx)
        self.assertIsInstance(f, filters.PandaQ)
        filtered = self.df[f.filter]
        n = filtered.shape[0]
        self.assertEqual(n, 2)




from noofa.tests.base import NoofaTest
from noofa.builders.builders import ReportBuilder


class TestBuilder(NoofaTest):
    """
    Тесты формирователя отчётов.
    """
    def setUp(self):
        self.sources = _sources_conf()
        self.queries = _queries_conf()
        self.dataframes = _dataframes_conf()
        self.tables = _tables_conf()
        self.figures = _figures_conf()
        self.conf = {
            'data_config': {
                'sources': self.sources,
                'queries': self.queries,
                'dataframes': self.dataframes,
            },
            'components_config': {**self.tables, **self.figures},
        }

    def test_init(self):
        success = True
        try:
            rb = ReportBuilder(**self.conf)
        except:
            success = False
        self.assertTrue(success)


def _sources_conf():
    return {
        'src1': {
            'id': 'src1',
            'type': 'postgres',
            'name': 'test_source',
            'from': 'json',
            'value': {
                'host': '',
                'port': 9999,
                'dbname': '',
                'user': '',
                'password': '',
            },
        },
        'src2': {
            'id': 'src2',
            'type': 'postgres',
            'name': 'test_source2',
            'from': 'conn_str',
            'value': 'host=1;port=1;dbname=1;user=1;password=1',
        },
    }

def _queries_conf():
    return {
        'q1': {
            'id': 'q1',
            'name': 'test_query1',
            'from': 'json',
            'source': 'src1',
            'value': {
                'base': 'table1',
                'tables': ['table1'],
                'filters': [
                    {
                        'is_complex': False,
                        'table': 'table1',
                        'field': 'col1',
                        'op': '>',
                        'value': 1,
                    }
                ],
            },
        },
        'q2': {
            'id': 'q2',
            'name': 'test_query2',
            'from': 'expression',
            'source': 'src1',
            'value': 'sql_select("a")',
        },
    }

def _dataframes_conf():
    return {
        'df1': {
            'id': 'df1',
            'name': 'test_df1',
            'base': {
                'type': 'query',
                'source': 'src1',
                'value': 'q1',
            },
        },
        'df2': {
            'id': 'df2',
            'name': 'test_df2',
            'base': {
                'type': 'source',
                'source': '',
                'value': 'src1',
            },
        },
        'df3': {
            'id': 'df3',
            'name': 'test_df3',
            'base': {
                'type': 'expression',
                'value': 'dataframe()',
            },
        },
    }

def _tables_conf():
    return {
        'table1': {
            'id': 'table1',
            'type': 'table',
            'base': {
                'from': 'dataframe',
                'value': 'df1',
            },
            'title_text': 'Table',
            'aliases': {'b': 'B'},
            'exclude': ['c'],
        },
        'pivottable1': {
            'id': 'pivottable1',
            'type': 'pivot_table',
            'base': {
                'from': 'dataframe',
                'value': 'df1',
            },
            'title_text': 'Pivot table',
            'pivot_conf': {
                'index': ['d'],
                'columns': [],
                'aggfunc': {'c': ['sum']},
                'values': {},
                'fill_value': None,
            },
        },
    }

def _figures_conf():
    return {
        'fig1': {
            'id': 'fig1',
            'type': 'figure',
            'base': {
                'from': 'dataframe',
                'value': {'dataframe': 'df1'},
                'x': 'a',
                'y': 'b',
            },
            'figure_type': 'line',
        },
        'line1': {
            'id': 'line1',
            'type': 'figure',
            'base': {
                'from': 'grouped',
                'value': {'df_from': 'expression', 'dataframe': 'df1'},
                'x': 'a',
                'y': 'b',
                'line_group': 'd',
            },
            'figure_type': 'line',
        },
        'line2': {
            'id': 'line2',
            'type': 'figure',
            'base': {
                'from': 'grouped',
                'value': {'df_from': 'dataframe', 'dataframe': 'df1'},
                'x': 'a',
                'y': 'b',
                'line_group': 'd',
            },
            'figure_type': 'line',
        },
        'line3': {
            'id': 'line3',
            'type': 'figure',
            'base': {
                'from': 'list',
                'value': [
                    {
                        'name': 'l1',
                        'x_from': 'expression',
                        'y_from': 'expression',
                        'x': 'df1',
                        'y': 'df1',
                    },
                ],
            },
            'figure_type': 'line',
        },
        'line4': {
            'id': 'line4',
            'type': 'figure',
            'base': {
                'from': 'list',
                'value': [
                    {
                        'name': 'l1',
                        'x_from': 'column',
                        'y_from': 'column',
                        'x': {'df_from': 'expression', 'dataframe': 'df1', 'column': 'a'},
                        'y': {'df_from': 'dataframe', 'dataframe': 'df1', 'column': 'b'},
                    },
                ],
            },
            'figure_type': 'line',
        },
        'bar1': {
            'id': 'bar1',
            'type': 'figure',
            'base': {
                'from': 'dataframe',
                'value': {'dataframe': 'df1'},
                'x': 'a',
                'y': 'b',
                'barmode': 'stack',
            },
            'figure_type': 'bar',
        },
        'bar2': {
            'id': 'bar2',
            'type': 'figure',
            'base': {
                'from': 'list',
                'value': [
                    {
                        'name': 'b1',
                        'x_from': 'column',
                        'y_from': 'column',
                        'x': {'df_from': 'expression', 'dataframe': 'df1', 'column': 'a'},
                        'y': {'df_from': 'dataframe', 'dataframe': 'df1', 'column': 'b'},
                    },
                ],
                'barmode': 'stack',
            },
            'figure_type': 'bar',
        },
        'bar3': {
            'id': 'bar3',
            'type': 'figure',
            'base': {
                'from': 'agg',
                'value': {
                    'dataframe': 'anything',
                    'groupby': ['d'],
                    'func': 'sum',
                    'on': 'a',
                },
                'barmode': 'stack',
            },
            'figure_type': 'bar',
        },
        'pie1': {
            'id': 'pie1',
            'type': 'figure',
            'base': {
                'from': 'dataframe',
                'value': {
                    'dataframe': 'df1',
                },
                'names': 'd',
                'values': 'a',
            },
            'figure_type': 'pie',
        },
        'pie2': {
            'id': 'pie2',
            'type': 'figure',
            'base': {
                'from': 'list',
                'value': [
                    {'name': 'a1', 'value': 'some_expression'},
                    {'name': 'a2', 'value': 'some_other_expression'},
                ],
            },
            'figure_type': 'pie',
        },
        'pie3': {
            'id': 'pie3',
            'type': 'figure',
            'base': {
                'from': 'agg',
                'value': {
                    'dataframe': 'anything',
                    'groupby': ['d'],
                    'func': 'sum',
                    'on': 'a',
                },
            },
            'figure_type': 'pie',
        },
    }



from pandas import DataFrame

from noofa.tests.base import NoofaTest
from noofa.core.dataframes import panda_builder as pb


class TestPandaBuilder(NoofaTest):
    """
    Тестирование функций работы с дф.
    """
    @classmethod
    def setUpClass(cls):
        cls.df2 = DataFrame([
            {'a': 4, 'c': 'Hello'},
            {'a': 2, 'c': 'hell'},
            {'a': 1, 'c': 'ello'},
            {'a': 3, 'c': 'ohell'},
            {'a': 2, 'c': 'hellno'},
            {'a': 2, 'c': 'holle'},
        ])
        cls.df_na = DataFrame([
            {'a': 4, 'c': None},
            {'a': 2, 'c': 2.5},
            {'a': 1, 'c': 3.14},
            {'a': 3, 'c': 2.7},
            {'a': 2, 'c': None},
            {'a': 2, 'c': None},
        ])
        cls.bad_filter = {
            'col_name': 'a',
            'op': 'non_ex',
            'value': 1,
            'is_q': False,
        }
        cls.filter = {
            'col_name': 'a',
            'op': '>',
            'value': 1,
            'is_q': False,
        }

    def setUp(self):
        self.df1 = DataFrame([
            {'a': 1, 'b': 'Hello'},
            {'a': 2, 'b': 'hell'},
            {'a': 2, 'b': 'ello'},
            {'a': 3, 'b': 'ohell'},
            {'a': 3, 'b': 'hellno'},
            {'a': 3, 'b': 'holle'},
        ])

    def test_new_and_empty(self):
        self.assertIsInstance(pb.new(), DataFrame)
        self.assertIsInstance(pb.empty(), DataFrame)

    def test_join(self):
        df = pb.join(self.df1, self.df2, ['a', 'a'], 'inner')
        r, c = df.shape
        self.assertEqual([r, c], [10, 3])

    def test_union_bad(self):
        self.assertRaises(Exception, pb.union, [])

        df = pb.union([self.df1])
        r, c = df.shape
        r1, c1 = self.df1.shape
        self.assertEqual([r, c], [r1, c1])

    def test_union(self):
        df = pb.union([self.df1, self.df2])
        r1 = self.df1.shape[0]
        r2 = self.df2.shape[0]
        r = df.shape[0]
        self.assertEqual([r1 + r2, 3], [r, 3])

    def test_add_column(self):
        col = 'd'
        cols = self.df1.shape[1]
        df = pb.add_column(self.df1, col, 1)
        cols1 = df.shape[1]
        self.assertTrue(cols1 == cols + 1 and col in df)

        df = pb.add_column(self.df1, col, 1)
        cols2 = df.shape[1]
        self.assertTrue(cols1 == cols2 and col in df)

    def test_add_columns(self):
        to_add = {'d': 1, 'e': 2}
        cols = self.df1.shape[1]
        df = pb.add_columns(self.df1, to_add)
        cols1 = df.shape[1]
        self.assertEqual(cols1, cols + 2)

        for col in to_add.keys():
            self.assertIn(col, df)

    def test_rename(self):
        cols_to_rename = {'a': 'A', 'b': 'B'}
        df = pb.rename_columns(self.df1, cols_to_rename)
        for prev, new in cols_to_rename.items():
            self.assertNotIn(prev, df)
            self.assertIn(new, df)

    def test_drop_empty(self):
        cols = self.df1.shape[1]
        df = pb.drop_columns(self.df1, [])
        self.assertEqual(cols, df.shape[1])

    def test_drop_multiple(self):
        cols = self.df1.shape[1]
        to_drop = ['a', 'b']
        df = pb.drop_columns(self.df1, to_drop)
        self.assertEqual(cols - 2, df.shape[1])
        for col in to_drop:
            self.assertNotIn(col, df)

    def test_filter_bad(self):
        self.assertRaises(KeyError, pb.filter, self.df1, [self.bad_filter])

    def test_filter(self):
        df = pb.filter(self.df1, [self.filter])
        self.assertEqual(df.shape[0], 5)

    def test_astype_simple(self):
        col = 'a'
        df = pb.astype(self.df1, col, 'str')
        self.assertEqual(str(df[col].dtype), 'object')

    def test_astype_int_to_dt(self):
        col = 'a'
        df = pb.astype(self.df1, col, 'datetime')
        self.assertTrue(str(df[col].dtype).startswith('datetime64'))

    def test_astype_dt_to_int(self):
        col = 'a'
        df = pb.astype(self.df1, col, 'datetime')
        df = pb.astype(df, col, 'int')
        self.assertEqual(str(df[col].dtype), 'int64')

    def test_astype_dt_to_float(self):
        col = 'a'
        df = pb.astype(self.df1, col, 'datetime')
        df = pb.astype(df, col, 'float')
        self.assertEqual(str(df[col].dtype), 'float64')

    def test_fill_na(self):
        col = 'c'
        self.assertRaises(AssertionError,
            pb.fill_na, self.df_na, col, None)

        df = pb.fill_na(self.df_na, col, 1)
        self.assertFalse(df[col].isnull().values.any())

    def test_drop_na(self):
        df = pb.drop_na(self.df_na, 'c')
        self.assertEqual(df.shape[0], 3)

    def test_pivot_table_empty(self):
        self.assertIsInstance(
            pb.pivot_table(DataFrame(), aggfunc={}),
            DataFrame
        )
        self.assertIsInstance(
            pb.pivot_table(self.df1, aggfunc={}),
            DataFrame
        )

    def test_pivot_table(self):
        self.assertIsInstance(
            pb.pivot_table(self.df1, index='b', aggfunc={'a': ['sum']}),
            DataFrame
        )
